# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os.path
import numpy as np
from glob import glob
import abel
from scipy.linalg import inv

###############################################################################
#
#  Dasch two-point, three_point, and onion-peeling deconvolution
#    as described in Applied Optics 31, 1146 (1992), page 1147-8 sect. B & C.
#        https://www.osapublishing.org/ao/abstract.cfm?uri=ao-31-8-1146
#    see also discussion in PR #155  https://github.com/PyAbel/PyAbel/pull/155
#
# 2016-03-25 Dan Hickstein - one line Abel transform
# 2016-03-24 Steve Gibson - Python code framework
# 2015-12-29 Dhrubajyoti Das - original three_point code and
#                              highlighting the Dasch paper,see issue #61
#                              https://github.com/PyAbel/PyAbel/issues/61
#
###############################################################################

_dasch_parameter_docstring = \
    """
    The dasch_method deconvolution method.

    C. J. Dasch,
    "One-dimensional tomography: a comparison of Abel, onion-peeling, and
    filtered backprojection methods",
    `Appl. Opt. 31, 1146–1152 (1992)
    <https://doi.org/10.1364/AO.31.001146>`__.

    Parameters
    ----------
    IM : 1D or 2D numpy array
        right-side half-image (or quadrant)

    basis_dir: str or None
        path to the directory for saving / loading the "dasch_method"
        deconvolution operator array. Here, called ``basis_dir`` for
        consistency with the other true basis methods. Use ``''`` for the
        default directory. If ``None``, the operator array will not be loaded
        from or saved to disk.

    dr : float
        sampling size (=1 for pixel images), used for Jacobian scaling.
        The resulting inverse transform is simply scaled by 1/dr.

    direction : str
        only the `direction="inverse"` transform is currently implemented

    verbose : bool
        trace printing


    Returns
    -------
    inv_IM: 1D or 2D numpy array
        the "dasch_method" inverse Abel transformed half-image

    """

# cache deconvolution operator array
_D = None
_method = None
_source = None   # 'cache', 'generated', or 'file', for unit testing


def two_point_transform(IM, basis_dir='', dr=1, direction="inverse",
                        verbose=False):
    return _dasch_transform(IM, basis_dir=basis_dir, dr=dr,
                            direction=direction, method="two_point",
                            verbose=verbose)


def three_point_transform(IM, basis_dir='', dr=1, direction="inverse",
                          verbose=False):
    return _dasch_transform(IM, basis_dir=basis_dir, dr=dr,
                            direction=direction, method="three_point",
                            verbose=verbose)


def onion_peeling_transform(IM, basis_dir='', dr=1, direction="inverse",
                            verbose=False):
    return _dasch_transform(IM, basis_dir=basis_dir, dr=dr,
                            direction=direction, method="onion_peeling",
                            verbose=verbose)


two_point_transform.__doc__ =\
            _dasch_parameter_docstring.replace("dasch_method", "two-point")
three_point_transform.__doc__ =\
            _dasch_parameter_docstring.replace("dasch_method", "three-point")
onion_peeling_transform.__doc__ =\
            _dasch_parameter_docstring.replace("dasch_method", "onion-peeling")


def _dasch_transform(IM, basis_dir='', dr=1, direction="inverse",
                     method="three_point", verbose=False):
    global _D, _method

    if direction != 'inverse':
        raise ValueError('Forward "two_point" transform not implemented')

    # make sure that the data has 2D shape
    IM = np.atleast_2d(IM)

    rows, cols = IM.shape

    if cols < 2 and method == "two_point":
        raise ValueError('"two_point" requires image width (cols) > 2')

    if cols < 3 and method == "three_point":
        raise ValueError('"three_point" requires image width (cols) > 3')

    _D = get_bs_cached(method, cols, basis_dir=basis_dir, verbose=verbose)

    inv_IM = dasch_transform(IM, _D)

    if rows == 1:
        inv_IM = inv_IM[0]  # flatten array

    return inv_IM/dr


def dasch_transform(IM, D):
    """Inverse Abel transform using the given deconvolution D-operator array.

    Parameters
    ----------
    IM : 2D numpy array
        image data

    D : 2D numpy array
        deconvolution operator array, of shape (cols, cols)

    Returns
    -------
    inv_IM : 2D numpy array
        inverse Abel transform according to deconvolution operator D
    """

    # one-line Abel transform - dot product of each row of IM with D
    return np.tensordot(IM, D, axes=(1, 1))


def _bs_two_point(cols):
    """deconvolution function for two_point.

    Parameters
    ----------
    cols : int
        width of the image
    """

    # function Eq. (9)  for j >= i
    def J(i, j):
        return np.log((np.sqrt((j + 1)**2 - i**2) + j + 1) /
                      (np.sqrt(j**2 - i**2) + j))/np.pi

    # Eq. (8, 9) D-operator basis, is 0 for j < i
    D = np.zeros((cols, cols))

    # diagonal i == j
    Ii, Jj = np.diag_indices(cols)
    Ii = Ii[1:]  # exclude special case i=j=0
    Jj = Jj[1:]
    D[Ii, Jj] = J(Ii, Jj)

    # upper triangle j > i
    Iu, Ju = np.triu_indices(cols, k=1)
    Iu = Iu[1:]  # exclude special case [0, 1]
    Ju = Ju[1:]
    D[Iu, Ju] = J(Iu, Ju) - J(Iu, Ju-1)

    # special cases
    D[0, 1] = J(0, 1) - 2/np.pi
    D[0, 0] = 2/np.pi

    return D


def _bs_three_point(cols):
    """deconvolution function for three_point.

    Parameters
    ----------
    cols : int
        width of the image
    """

    # function Eq. (7)  for j >= i
    def I0diag(i, j):
        return np.log((np.sqrt((2*j+1)**2-4*i**2) + 2*j+1)/(2*j))/(2*np.pi)

    # j > i
    def I0(i, j):
        return np.log(((np.sqrt((2*j + 1)**2 - 4*i**2) + 2*j + 1)) /
                      (np.sqrt((2*j - 1)**2 - 4*i**2) + 2*j - 1))/(2*np.pi)

    # i = j  NB minus -2I_ij typo in Dasch paper
    def I1diag(i, j):
        return np.sqrt((2*j+1)**2 - 4*i**2)/(2*np.pi) - 2*j*I0diag(i, j)

    # j > i
    def I1(i, j):
        return (np.sqrt((2*j+1)**2 - 4*i**2) -
                np.sqrt((2*j-1)**2 - 4*i**2))/(2*np.pi) - 2*j*I0(i, j)

    D = np.zeros((cols, cols))

    # matrix indices ------------------
    # i = j
    I, J = np.diag_indices(cols)
    I = I[1:]
    J = J[1:]  # drop special cases (0,0), (0,1)

    # j = i - 1
    Ib, Jb = I, J-1

    # j = i + 1
    Iu, Ju = I-1, J
    Iu = Iu[1:]  # drop special case (0, 1)
    Ju = Ju[1:]

    # j > i + 1
    Iut, Jut = np.triu_indices(cols, k=2)
    Iut = Iut[1:]  # drop special case (0, 2)
    Jut = Jut[1:]

    # D operator matrix ------------------
    # j = i - 1
    D[Ib, Jb] = I0diag(Ib, Jb+1) - I1diag(Ib, Jb+1)

    # j = i
    D[I, J] = I0(I, J+1) - I1(I, J+1) + 2*I1diag(I, J)

    # j = i + 1
    D[Iu, Ju] = I0(Iu, Ju+1) - I1(Iu, Ju+1) + 2*I1(Iu, Ju) -\
                I0diag(Iu, Ju-1) - I1diag(Iu, Ju-1)

    # j > i + 1
    D[Iut, Jut] = I0(Iut, Jut+1) - I1(Iut, Jut+1) + 2*I1(Iut, Jut) -\
                  I0(Iut, Jut-1) - I1(Iut, Jut-1)

    # special cases (that switch between I0, I1 cases)
    D[0, 2] = I0(0, 3) - I1(0, 3) + 2*I1(0, 2) - I0(0, 1) - I1(0, 1)
    D[0, 1] = I0(0, 2) - I1(0, 2) + 2*I1(0, 1) - 1/np.pi
    D[0, 0] = I0(0, 1) - I1(0, 1) + 1/np.pi

    return D


def _bs_onion_peeling(cols):
    """deconvolution function for onion_peeling.

    Parameters
    ----------
    cols : int
        width of the image
    """

    # weight matrix
    W = np.zeros((cols, cols))

    # diagonal elements i = j, Eq. (11)
    I, J = np.diag_indices(cols)
    W[I, J] = np.sqrt((2*J+1)**2 - 4*I**2)

    # upper triangle j > i,  Eq. (11)
    Iu, Ju = np.triu_indices(cols, k=1)
    W[Iu, Ju] = np.sqrt((2*Ju + 1)**2 - 4*Iu**2) -\
                np.sqrt((2*Ju - 1)**2 - 4*Iu**2)

    # operator used in Eq. (1)
    D = inv(W)

    return D


def get_bs_cached(method, cols, basis_dir='', verbose=False):
    """Load Dasch method deconvolution operator array from cache, or disk.
    Generate and store if not available.

    Checks whether ``method`` deconvolution array has been previously
    calculated, or whether the file ``{method}_basis_{cols}.npy`` is
    present in `basis_dir`.

    Either, assign, read, or generate the deconvolution array
    (saving it to file).


    Parameters
    ----------
    method : str
        Abel transform method ``onion_peeling``, ``three_point``, or
        ``two_point``

    cols : int
        width of image

    basis_dir : str or None
        path to the directory for saving or loading the deconvolution array.
        Use ``''`` for the default directory.
        For ``None``, do not load or save the deconvolution operator array

    verbose: boolean
        print information (mainly for debugging purposes)

    Returns
    -------
    D: numpy 2D array of shape (cols, cols)
       deconvolution operator array for the associated method

    file.npy: file
       saves `D`, the deconvolution array to file name:
       ``{method}_basis_{cols}.npy``

    """

    global _D, _method, _source

    # check whether the deconvolution operator array is cached
    if _D is not None:
        if _D.shape[0] >= cols and _method == method:
            if verbose:
                print('Using memory cached deconvolution operator array,'
                      ' shape {}'.format(_D.shape))
            _source = 'cache'
            return _D[:cols, :cols]  # sliced to correct size

    D_name = "{}_basis_{}.npy".format(method, cols)
    D_generator = {
        "onion_peeling": abel.dasch._bs_onion_peeling,
        "three_point": abel.dasch._bs_three_point,
        "two_point": abel.dasch._bs_two_point
    }

    _method = method

    if basis_dir == '':
        basis_dir = abel.transform.get_basis_dir(make=True)

    # read deconvolution operator array if available
    if basis_dir is not None:
        path_to_basis_files = os.path.join(basis_dir, method+'_basis*')
        basis_files = glob(path_to_basis_files)
        for bf in basis_files:
            if int(bf.split('_')[-1].split('.')[0]) >= cols:
                # relies on file order
                if verbose:
                    print("Loading deconvolution operator array from"
                          " file {:s}".format(bf))
                # slice to size
                _D = np.load(bf)[:cols, :cols]
                _source = 'file'
                return _D

    if verbose:
        print("A suitable deconvolution array for '{:s}' was not found.".
              format(method))
        print("A new array will be generated.")

    _D = D_generator[method](cols)
    _source = 'generated'

    if basis_dir is not None:
        path_to_basis_file = os.path.join(basis_dir, D_name)
        np.save(path_to_basis_file, _D)
        if verbose:
            print("\ndeconvolution operator array saved to '{:s}"
                  .format(path_to_basis_file))

    return _D


def cache_cleanup():
    """
    Utility function.

    Frees the memory caches created by :func:`get_bs_cached`.
    This is usually pointless, but might be required after working
    with very large images, if more RAM is needed for further tasks.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """

    global _D, _method, _source

    _D = None
    _method = None
    _source = None


def basis_dir_cleanup(method, basis_dir=''):
    """
    Utility function.

    Deletes deconvolution operator arrays saved on disk.

    Parameters
    ----------
    method : str
        Abel transform method ``'onion_peeling'``, ``'three_point'``, or
        ``'two_point'``
    basis_dir : str or None
        absolute or relative path to the directory with saved deconvolution
        operator arrays. Use ``''`` for the default directory. ``None`` does
        nothing.

    Returns
    -------
    None
    """
    if basis_dir == '':
        basis_dir = abel.transform.get_basis_dir(make=False)

    if basis_dir is None:
        return

    if method not in ['onion_peeling', 'three_point', 'two_point']:
        raise ValueError('Incorrect method "{}"!'.format(method))

    files = glob(os.path.join(basis_dir, method + '_basis_*.npy'))
    for fname in files:
        os.remove(fname)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import numpy as np
from numpy.testing import assert_allclose, assert_equal

from abel.tools.analytical import StepAnalytical, GaussianAnalytical
from abel.benchmark import absolute_ratio_benchmark, is_symmetric


def assert_allclose_msg(x, y, message):
    assert np.allclose(x, y), message


def test_symmetry_function():
    # Check the consistency of abel.benchmark.is_symmetric

    # checking all combination of odd and even images
    for n, m in [(11, 11),
                 (10, 10),
                 (10, 12),
                 (11, 12),
                 (10, 13),
                 (11, 13)
                 ]:
        x = np.linspace(-np.pi, np.pi, n)
        y = np.linspace(-np.pi, np.pi, m)

        XX, YY = np.meshgrid(x, y)


        f1 = np.cos(XX)*np.cos(YY)
        assert_equal( is_symmetric(f1, i_sym=True, j_sym=True), True,
              'Checking f1 function for polar symmetry n={},m={}'.format(n,m))

        f2 = np.cos(XX)*np.sin(YY)
        assert_equal( is_symmetric(f2, i_sym=True, j_sym=False), True,
              'Checking f2 function for i symmetry n={},m={}'.format(n,m))

        f3 = np.sin(XX)*np.cos(YY)
        assert_equal( is_symmetric(f3, i_sym=False, j_sym=True), True,
              'Checking f3 function for j symmetry n={},m={}'.format(n,m))

        assert_equal( is_symmetric(f3, i_sym=True, j_sym=False), False,
              'Checking that f3 function does not have i symmetry n={},m={}'.format(n,m))

    for n in [10, 11]:
        x = np.linspace(-np.pi, np.pi, n)
        f1 = np.cos(x)
        assert_equal( is_symmetric(f1, i_sym=True, j_sym=False), True,\
              'Checking f1 function for symmetry in 1D n={},m={}'.format(n,m))


def test_absolute_ratio_benchmark():
    # Mostly sanity to check that Analytical functions don't have typos
    # or syntax errors
    n = 501
    r_max = 50

    for symmetric in [True, False]:
        for Backend, options in [(StepAnalytical, dict(A0=10.0, r1=6.0,
                                            r2=14.0, ratio_valid_step=1.0)),
                                 (GaussianAnalytical, dict(sigma=2))]:

            ref = Backend(n, r_max, symmetric=symmetric, **options)
            ratio = absolute_ratio_benchmark(ref, ref.func)

            backend_name = type(ref).__name__


            assert_allclose_msg(ratio.mean(), 1.0,
                    "Sanity test, sym={}: {}: ratio == 1.0".format(symmetric, backend_name))
            assert_allclose_msg(ratio.std(), 0.0,
                    "Sanity test, sym={}: {}: std == 0.0".format(symmetric, backend_name))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from .tools.math import gradient

try:
    from .lib.direct import _cabel_direct_integral
    cython_ext = True
except (ImportError, UnicodeDecodeError):
    cython_ext = False

###########################################################################
# direct - calculation of forward and inverse Abel transforms by direct
# numerical integration
#
# Roman Yurchak - Laboratoire LULI, Ecole Polytechnique/CNRS/CEA, France
# 07.2018: DH fixed the correction for the case where r[0] = 0
# 03.2018: DH changed the default grid from 0.5, 1.5 ... to 0, 1, 2.
# 01.2018: DH dhanged the integration method to trapz
# 12.2015: RY Added a pure python implementation
# 11.2015: RY moved to PyAbel, added more unit tests, reorganized code base
#    2012: RY first implementation in hedp.math.abel
###########################################################################


def _construct_r_grid(n, dr=None, r=None):
    """ Internal function to construct a 1D spatial grid """
    if dr is None and r is None:
        # default value, we don't care about the scaling
        # since the mesh size was not provided
        dr = 1.0

    if dr is not None and r is not None:
        raise ValueError('Both r and dr input parameters cannot be specified \
                            at the same time')
    elif dr is None and r is not None:
        if r.ndim != 1 or r.shape[0] != n:
            raise ValueError('The input parameter r should be a 1D array'
                             'of shape = ({},), got shape = {}'.format(
                                                                n, r.shape))
        # not so sure about this, needs verification -RY
        dr = np.gradient(r)

    else:
        if isinstance(dr, np.ndarray):
            raise NotImplementedError
        r = (np.arange(n))*dr
    return r, dr


def direct_transform(fr, dr=None, r=None, direction='inverse',
                     derivative=gradient, int_func=np.trapz,
                     correction=True, backend='C', **kwargs):
    """
    This algorithm performs a direct computation of the Abel transform
    integrals. When correction=False, the pixel at the lower bound of the
    integral (where y=r) is skipped, which causes a systematic error in the
    Abel transform. However, if correction=True is used, then an analytical
    transform transform is applied to this pixel, which makes the approximation
    that the function is linear across this pixel. With correction=True, the
    Direct method produces reasonable results.

    The Direct method is implemented in both Python and a compiled C version
    using Cython, which is much faster. The implementation can be selected using
    the backend argument. If the C-backend is not available, you must re-install
    PyAbel with Numpy, Cython, and a C-compiler already installed.

    By default, integration at all other pixels is performed using the
    Trapezoidal rule.

    Parameters
    ----------

    fr : 1d or 2d numpy array
        input array to which direct/inverse Abel transform will be applied.
        For a 2d array, the first dimension is assumed to be the z axis and
        the second the r axis.
    dr : float
        spatial mesh resolution (optional, default to 1.0)
    r : 1D ndarray
        the spatial mesh (optional). Unusually, direct_transform should, in
        principle, be able to handle non-uniform data. However, this has not
        been regorously tested.
    direction : string
        Determines if a forward or inverse Abel transform will be applied.
        can be 'forward' or 'inverse'.
    derivative : callable
        a function that can return the derivative of the fr array
        with respect to r. (only used in the inverse Abel transform).
    int_func : function
        This function is used to complete the integration. It should resemble
        np.trapz, in that it must be callable using axis=, x=, and dx=
        keyword arguments.
    correction : boolean
        If False the pixel where the weighting function has a singular value
        (where r==y) is simply skipped, causing a systematic under-estimation
        of the Abel transform.
        If True, integration near the singular value is performed analytically,
        by assuming that the data is linear across that pixel. The accuracy
        of this approximation will depend on how the data is sampled.
    backend : string
        There are currently two implementations of the Direct transform,
        one in pure Python and one in Cython. The backend paremeter selects
        which method is used. The Cython code is converted to C and compiled,
        so this is faster.
        Can be 'C' or 'python' (case insensitive).
        'C' is the default, but 'python' will be used
        if the C-library is not available.

    Returns
    -------
    out : 1d or 2d numpy array of the same shape as fr
        with either the direct or the inverse abel transform.
    """

    backend = backend.lower()
    if backend not in ['c', 'python']:
        raise ValueError
    f = np.atleast_2d(fr.copy())

    r, dr = _construct_r_grid(f.shape[1], dr=dr, r=r)

    if direction == "inverse":
        f = derivative(f)/dr
        f *= - 1./np.pi
    else:
        f *= 2*r[None, :]

    if backend == 'c':
        if not cython_ext:
            print('Warning: Cython extensions were not built, \
                    the C backend is not available!')
            print('Falling back to a pure Python backend...')
            backend = 'python'
        elif not is_uniform_sampling(r):
            print('Warning: non uniform sampling is currently not \
                    supported by the C backend!')
            print('Falling back to a pure Python backend...')
            backend = 'python'

    f = np.asarray(f, order='C', dtype='float64')
    if backend == 'c':
        out = _cabel_direct_integral(f, r, int(correction))
    else:
        out = _pyabel_direct_integral(f, r, int(correction), int_func)

    if f.shape[0] == 1:
        return out[0]
    else:
        return out


def _pyabel_direct_integral(f, r, correction, int_func=np.trapz):
    """
    Calculation of the integral  used in Abel transform
    (both direct and inverse).
             ∞
            ⌠
            ⎮      f(r)
            ⎮ ────────────── dr
            ⎮    ___________
            ⎮   ╱  2   2
            ⎮ ╲╱  y - r
            ⌡
            y
    Returns:
    --------
    np.array: of the same shape as f with the integral evaluated at r

    """
    if correction not in [0, 1]:
        raise ValueError

    if is_uniform_sampling(r):
        int_opts = {'dx': abs(r[1] - r[0])}
    else:
        int_opts = {'x': r}

    out = np.zeros(f.shape)
    R, Y = np.meshgrid(r, r, indexing='ij')

    i_vect = np.arange(len(r), dtype=int)
    II, JJ = np.meshgrid(i_vect, i_vect, indexing='ij')
    mask = (II < JJ)

    I_sqrt = np.zeros(R.shape)
    I_sqrt[mask] = np.sqrt((Y**2 - R**2)[mask])

    I_isqrt = np.zeros(R.shape)
    I_isqrt[mask] = 1./I_sqrt[mask]

    # create a mask that just shows the first two points of the integral
    mask2 = ((II > JJ-2) & (II < JJ+1))

    for i, row in enumerate(f):  # loop over rows (z)
        P = row[None, :] * I_isqrt  # set up the integral
        out[i, :] = int_func(P, axis=1, **int_opts)  # take the integral

        # correct for the extra triangle at the start of the integral
        out[i, :] = out[i, :] - 0.5*int_func(P*mask2, axis=1, **int_opts)

    """
    Compute the correction. Here we apply an
    analytical integration of the cell with the singular value,
    assuming a piecewise linear behaviour of the data.
    The analytical abel transform for this trapezoid is:
    c0*acosh(r1/y) - c_r*y*acosh(r1/y) + c_r*sqrt(r1**2 - y**2)
    see: https://github.com/luli/hedp/blob/master/hedp/math/abel.py#L87-L104
    """
    if correction == 1:

        # precompute a few variables outside the loop:
        f_r = (f[:, 1:] - f[:, :-1])/np.diff(r)[None, :]
        isqrt = I_sqrt[II+1 == JJ]

        if r[0] < r[1]*1e-8:  # special case for r[0] = 0
            ratio = np.append(np.cosh(1), r[2:]/r[1:-1])
        else:
            ratio = r[1:]/r[:-1]

        acr = np.arccosh(ratio)

        for i, row in enumerate(f):  # loop over rows (z)
            out[i, :-1] += isqrt*f_r[i] + acr*(row[:-1] - f_r[i]*r[:-1])

    return out


def is_uniform_sampling(r):
    """
    Returns True if the array is uniformly spaced to within 1e-13.
    Otherwise False.
    """
    dr = np.diff(r)
    ddr = np.diff(dr)
    return np.allclose(ddr, 0, atol=1e-13)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import absolute_import
import numpy as np
import abel
import scipy.constants as const
import scipy.interpolate

# This file includes functions that have a known analytical Abel transform.
# They are used in unit testing and for comparing different Abel
# impementations.


class BaseAnalytical(object):
    def __init__(self, n, r_max, symmetric=True, **args):
        """
        This is the base class for functions that have a known Abel transform.
        Every such class should expose the following public attributes:
          - self.r: vector of positions along the r axis
          - self.func: the values of the original function
                        (same shape as self.r)
          - self.abel: the values of the Abel transform (same shape as self.r)
          - self.mask_valid: mask of the r interval where the function is
            well smoothed/well behaved (no known artefacts in the inverse
            Abel reconstuction), typically excluding the origin, the domain
            boundaries, and function discontinuities, that can be used for
            unit testing.

        See GaussianAnalytical for a concrete example.

        Parameters
        -------a--
        n : int
            number of points along the r axis

        r_max: float
            maximum r interval

        symmetric: boolean
            if True, the r interval is [-r_max, r_max] (and n should be odd),
            otherwise, the r interval is [0, r_max]
        """
        self.n = n
        self.r_max = r_max

        assert r_max > 0

        if symmetric:
            self.r = np.linspace(-r_max, r_max, n)
        else:
            self.r = np.linspace(0, r_max, n)

        self.dr = np.diff(self.r)[0]


class StepAnalytical(BaseAnalytical):
    """
    Define a symmetric step function and calculate its analytical
    Abel transform. See :doc:`examples/example_basex_step.py
    <example_basex_step>`.

    Parameters
    ----------
    n : int
        number of points along the r axis
    r_max : float
        range of the r interval
    symmetric : boolean
        if True, the r interval is [-r_max, r_max] (and n should be odd),
        otherwise the r interval is [0, r_max]
    r1, r2 : float
        bounds of the step function if r > 0
        (symmetric function is constructed for r < 0)
    A0: float
        height of the step
    ratio_valid_step: float
        in the benchmark take only the central ratio*100% of the step
        (exclude possible artefacts on the edges)
    """
    # see https://github.com/PyAbel/PyAbel/pull/16

    def __init__(self, n, r_max, r1, r2, A0=1.0,
                 ratio_valid_step=1.0, symmetric=True):
        super(StepAnalytical, self).__init__(n, r_max, symmetric)

        self.r1, self.r2 = r1, r2
        self.A0 = A0

        mask = np.abs(np.abs(self.r) - 0.5*(r1 + r2)) < 0.5*(r2 - r1)

        fr = np.zeros(self.r.shape)
        fr[mask] = A0

        self.func = fr

        self.abel = self.sym_abel_step_1d(self.r, self.A0, self.r1, self.r2)[0]

        # exclude the region near the discontinuity
        self.mask_valid = np.abs(
            np.abs(self.r) - 0.5*(r1 + r2)) < ratio_valid_step*0.5*(r2 - r1)

    def abel_step_analytical(self, r, A0, r0, r1):
        """
        Forward Abel transform of a step function located between r0 and r1,
        with a height A0

        ::

            A0 +                  +-------------+
               |                  |             |
               |                  |             |
             0 | -----------------+             +-------------
               +------------------+-------------+------------>
               0                  r0            r1           r axis


        Parameters
        ----------
        r1 : 1D array
            vecor of positions along the r axis. Must start with 0.
        r0, r1 : float
            positions of the step along the r axis
        A0 : float or 1D array
            height of the step. If 1D array, the height can be variable
            along the z axis

        Returns
        -------
            1D array, if A0 is a float, a 2D array otherwise
        """

        if np.all(r[[0, -1]]):
            raise ValueError('The vector of r coordinates must start with 0.0')

        F_1d = np.zeros(r.shape)
        mask = (r >= r0)*(r < r1)
        F_1d[mask] = 2*np.sqrt(r1**2 - r[mask]**2)

        mask = r < r0
        F_1d[mask] = 2*np.sqrt(r1**2 - r[mask]**2) -\
                     2*np.sqrt(r0**2 - r[mask]**2)

        A0 = np.atleast_1d(A0)
        if A0.ndim == 1:
            A0 = A0[:, np.newaxis]

        return F_1d[np.newaxis, :]*A0

    def sym_abel_step_1d(self, r, A0, r0, r1):
        """
        Produces a symmetrical analytical transform of a 1D step
        """
        A0 = np.atleast_1d(A0)
        d = np.empty(A0.shape + r.shape)
        A0 = A0[np.newaxis, :]

        for sens, mask in enumerate([r >= 0, r <= 0]):
            d[:, mask] = self.abel_step_analytical(np.abs(r[mask]), A0, r0, r1)

        return d


class Polynomial(BaseAnalytical):
    r"""
    Define a polynomial function and calculate its analytical
    Abel transform.

    (See :ref:`Polynomials` for details and examples.)

    Parameters
    ----------
    n : int
        number of points along the *r* axis
    r_max : float
        range of the *r* interval
    symmetric : boolean
        if ``True``, the *r* interval is [−\ **r_max**, +\ **r_max**]
        (and **n** should be odd),
        otherwise the *r* interval is [0, **r_max**]
    r_1, r_2 : float
        *r* bounds of the polynomial function if *r* > 0;
        outside [**r_1**, **r_2**] the function is set to zero
        (symmetric function is constructed for *r* < 0)
    c: numpy array
        polynomial coefficients in order of increasing degree:
        [c₀, c₁, c₂] means c₀ + c₁ *r* + c₂ *r*\ ²
    r_0 : float, optional
        origin shift: the polynomial is defined as
        c₀ + c₁ (*r* − **r_0**) + c₂ (*r* − **r_0**)² + ...
    s : float, optional
        *r* stretching factor (around **r_0**): the polynomial is defined as
        c₀ + c₁ (*r*/s) + c₂ (*r*/s)² + ...
    reduced : boolean, optional
        internally rescale the *r* range to [0, 1];
        useful to avoid floating-point overflows for high degrees
        at large *r* (and might improve numerical accuracy)
    """
    def __init__(self, n, r_max,
                 r_1, r_2, c, r_0=0.0, s=1.0, reduced=False,
                 symmetric=True):
        super(Polynomial, self).__init__(n, r_max, symmetric)

        # take r >= 0 part
        if symmetric:
            r = self.r[n//2:]
        else:
            r = self.r

        P = abel.tools.polynomial.Polynomial(r, r_1, r_2, c, r_0, s, reduced)
        self.func = P.func
        self.abel = P.abel

        # mirror to negative r, if needed
        if symmetric:
            self.func = np.hstack((self.func[:0:-1], self.func))
            self.abel = np.hstack((self.abel[:0:-1], self.abel))

        self.mask_valid = np.ones_like(self.func)


class PiecewisePolynomial(BaseAnalytical):
    r"""
    Define a piecewise polynomial function (sum of ``Polynomial``\ s)
    and calculate its analytical Abel transform.

    Parameters
    ----------
    n : int
        number of points along the *r* axis
    r_max : float
        range of the *r* interval
    symmetric : boolean
        if ``True``, the *r* interval is [−\ **r_max**, +\ **r_max**]
        (and **n** should be odd),
        otherwise the *r* interval is [0, **r_max**]
    ranges : iterable of unpackable
        (list of tuples of) polynomial parameters for each piece::

           [(r_1_1st, r_2_1st, c_1st),
            (r_1_2nd, r_2_2nd, c_2nd),
            ...
            (r_1_nth, r_2_nth, c_nth)]

        according to ``Polynomial`` conventions.
        All ranges are independent (may overlap and have gaps, may define
        polynomials of any degrees) and may include optional ``Polynomial``
        parameters
    """
    def __init__(self, n, r_max,
                 ranges,
                 symmetric=True):
        super(PiecewisePolynomial, self).__init__(n, r_max, symmetric)

        # take r >= 0 part
        if symmetric:
            r = self.r[n//2:]
        else:
            r = self.r

        P = abel.tools.polynomial.PiecewisePolynomial(r, ranges)
        self.func = P.func
        self.abel = P.abel

        # mirror to negative r, if needed
        if symmetric:
            self.func = np.hstack((self.func[:0:-1], self.func))
            self.abel = np.hstack((self.abel[:0:-1], self.abel))

        self.mask_valid = np.ones_like(self.func)


class GaussianAnalytical(BaseAnalytical):
    """
    Define a gaussian function and calculate its analytical
    Abel transform. See :doc:`examples/example_basex_gaussian.py
    <example_basex_gaussian>`.

    Parameters
    ----------
    n : int
        number of points along the r axis
    r_max : float
        range of the r interval
    symmetric : boolean
        if True, the r interval is [-r_max, r_max] (and n should be odd),
        otherwise, the r interval is [0, r_max]
    sigma : floats
        sigma parameter for the gaussian
    A0 : float
        amplitude of the gaussian
    ratio_valid_sigma : float
        in the benchmark take only the range
        0 < r < ration_valid_sigma * sigma
        (exclude possible artefacts on the axis and the possibly clipped tail)
    """
    # Source: http://mathworld.wolfram.com/AbelTransform.html

    def __init__(self, n, r_max, sigma=1.0, A0=1.0,
                 ratio_valid_sigma=2.0, symmetric=True):
        super(GaussianAnalytical, self).__init__(n, r_max, symmetric)

        self.sigma = sigma
        self.A0 = A0

        r = self.r

        self.func = A0*np.exp(-r**2/sigma**2)

        self.abel = sigma * np.sqrt(np.pi) * A0*np.exp(-r**2/sigma**2)

        # exclude the region near the discontinuity
        self.mask_valid = (np.abs(self.r) < ratio_valid_sigma*sigma) &\
                          (np.abs(self.r) > 0)


class TransformPair(BaseAnalytical):
    """**Abel-transform pair analytical functions**.

    **profiles 1–7**: Table 1 of
    G. C.-Y. Chan, Gary M. Hieftje,
    "Estimation of confidence intervals for radial emissivity and optimization
    of data treatment techniques in Abel inversion",
    `Spectrochimica Acta B 61, 31–41 (2006)
    <https://doi.org/10.1016/j.sab.2005.11.009>`_.

    See :mod:`abel.tools.transform_pairs`.


    Returns
    -------
    r : numpy array
        vector of positions along the r axis: `linspace(0, 1, n)`

    dr : float
        radial interval

    func : numpy array
        values of the original function (same shape as r)

    abel : numpy array
        values of the Abel transform (same shape as func)

    label : str
        name of the curve

    mask_valid : boolean array
        set all True. Used for unit tests

    """

    def __init__(self, n, profile=5):
        """Create Abel transform pair for given profile number.

        Parameters
        ----------
        n : int
            number of points along the r axis

        profile: int
            the profile number 1-7, see 'abel/tools/transform_pairs.py'

        """

        super(TransformPair, self).__init__(n, r_max=1, symmetric=False)

        # BaseAnalytical creates self.r = linspace(0, 1, n)
        # prevent divide by zero and NaN for r = 0, or 1,
        # slightly offset these values
        r = self.r.copy()
        r[0] = 1.0e-8
        r[-1] -= 1.0e-8

        if profile > 7:
            raise ValueError('Only 1-7 profiles: '
                             'see "abel/tools/transform_pairs.py"')

        self.label = 'profile{}'.format(profile)

        self.profile = getattr(abel.tools.transform_pairs, self.label)
        self.func, self.abel = self.profile(r)

        # function values to use for testing
        self.mask_valid = np.ones_like(self.func)


class SampleImage(BaseAnalytical):
    """
    Sample images, made up of Gaussian functions

    Parameters
    ----------
    n : integer
        image size n rows x n cols

    name : str
        one of "dribinski" or "Ominus"

    sigma : float
        Gaussian 1/e width (pixels)

    temperature : float
        for 'Ominus' only
        anion levels have Boltzmann population weight
        (2J+1) exp(-177.1 h c 100/k/temperature)

    Attributes
    ----------
    image : 2D numpy array
         image

    name : str
         sample image name
    """
    def __init__(self, n=361, name="dribinski", sigma=2, temperature=200):

        def _gauss(r, r0, sigma):
            return np.exp(-(r-r0)**2/sigma**2)

        def _dribinski(r, theta, sigma):
            """
            Sample test image used in the BASEX paper
            Rev. Sci. Instrum. 73, 2634 (2002), intensity function Eq. (16)
            (there are some missing negative exponents in the publication)

            9 Gaussian peaks with "width" = sigma (default: 2) and
            1 background Gaussian with "width" = 60
            ("width" is √2 std. dev. in pixels for default n = 361
             and scales proportionally to n)

            anisotropy: ß = −1  for cosθ term
                        ß = +2  for sinθ term
                        ß =  0  isotropic, no angular variation
            """

            sinetheta2 = np.sin(theta)**2
            cosinetheta2 = np.cos(theta)**2

            t0 = 7*_gauss(r, 10, sigma)*sinetheta2
            t1 = 3*_gauss(r, 15, sigma)
            t2 = 5*_gauss(r, 20, sigma)*cosinetheta2
            t3 = _gauss(r, 70, sigma)
            t4 = 2*_gauss(r, 85, sigma)*cosinetheta2
            t5 = _gauss(r, 100, sigma)*sinetheta2
            t6 = 2*_gauss(r, 145, sigma)*sinetheta2
            t7 = _gauss(r, 150, sigma)
            t8 = 3*_gauss(r, 155, sigma)*cosinetheta2
            t9 = 20*_gauss(r, 45, 60)  # background under t3 to t5

            return 2000*(t0+t1+t2) + 200*(t3+t4+t5) + 50*(t6+t7+t8) + t9

        def _Ominus(r, theta, sigma, boltzmann, rfact=1):
            """
            Simulate the photoelectron spectrum of O- photodetachment
            3PJ <- 2P3/2,1/2

            6 transitions, triplet neutral, and doublet anion

            """
            # positions based on 812.5 nm ANU O- PES
            t1 = _gauss(r, 341*rfact, sigma)  # 3P2 <- 2P3/2
            t2 = _gauss(r, 285*rfact, sigma)  # 3P1 <- 2P3/2
            t3 = _gauss(r, 257*rfact, sigma)  # 3P0 <- 2P3/2
            t4 = _gauss(r, 394*rfact, sigma)  # 3P2 <- 2P1/2
            t5 = _gauss(r, 348*rfact, sigma)  # 3P1 <- 2P1/2
            t6 = _gauss(r, 324*rfact, sigma)  # 3P0 <- 2P1/2

            # intensities = fine-structure known ratios
            # 2P1/2 transtions scaled by temperature-dependent Boltzmann factor
            t = t1 + 0.8*t2 + 0.36*t3 + (0.2*t4 + 0.36*t5 + 0.16*t6)*boltzmann

            # anisotropy
            sinetheta2 = np.sin(theta)**2*0.2 + 0.8

            return t*sinetheta2

        n2 = n//2
        self.name = name

        super(SampleImage, self).__init__(n, r_max=n2, symmetric=True)

        if name == 'dribinski':
            scale = 180*2/n
        elif name == 'Ominus':
            scale = 501*2/n

        X, Y = np.meshgrid(self.r*scale, self.r*scale)
        R, THETA = abel.tools.polar.cart2polar(X, Y)

        if self.name == "dribinski":
            self.image = _dribinski(R, THETA, sigma=sigma)
        elif self.name == "Ominus":
            boltzmann = np.exp(-177.1*const.h*const.c*100/const.k/temperature)\
                        / 2
            self.image = _Ominus(R, THETA, sigma=sigma, boltzmann=boltzmann)
        else:
            raise ValueError('Sample image name not recognized')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

import numpy as np
from numpy.testing import assert_allclose

import abel

def test_symmetry_get_put_quadrants(image_shape=(5,5), verbose=False):
    
    n, m = image_shape
    z = np.ones((n, m))*(-1)     # odd-image axes tagged with '-1'
    c2 = m//2 
    r2 = n//2 
    # tag each quadrant with its number
    z[:r2, -c2:] = 0
    z[:r2, :c2] = 1
    z[-r2:, :c2] = 2
    z[-r2:, -c2:] = 3 

    if verbose:
        print("test image of shape ", z.shape)
        if n % 2 or m % 2:
            print(" odd-size axes tagged with '-1'")
        print(z)

    q = abel.tools.symmetry.get_image_quadrants(z, reorient=True)

    if verbose:
        for i, qi in enumerate(q):
            print("\nreoriented quadrant Q{:d}".format(i))
            print(qi)

    r = abel.tools.symmetry.put_image_quadrants(q, original_image_shape=z.shape)

    if verbose:
        print("\nreassembled image, shape = ", r.shape)
        print(r)

    assert not (z-r).any()

"""
    q = abel.tools.symmetry.get_image_quadrants(z, reorient=True, 
            use_quadrants=(True, False, False, False), symmetry_axis=(0, 1)) 

    if verbose:
        for i, qi in enumerate(q):
            print("\nreoriented quadrant Q{:d}".format(i))
            print(qi)
"""


if __name__ == "__main__":
  test_symmetry_get_put_quadrants(image_shape=(5, 5), verbose=True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import abel
from . import basex
from . import dasch
from . import daun
from . import direct
from . import hansenlaw
from . import linbasex
from . import onion_bordas
from . import rbasex
from . import tools

from timeit import default_timer as timer
import itertools
import sys


def _ensure_list(x):
    """
    Convert the argument to a list (a scalar becomes a single-element list).
    """
    return [x] if np.ndim(x) == 0 else list(x)


def _roundsf(x, n):
    """
    Round to n significant digits
    """
    return float('{:.{p}g}'.format(x, p=n))


class Timent(object):
    """
    Helper class for measuring execution times.

    The constructor only initializes the timing-procedure parameters.
    Use the :py:meth:`.time` method to run it for particular functions.

    Parameters
    ----------
    skip : int
        number of "warm-up" iterations to perform before the measurements.
        Can be specified as a negative number, then ``abs(skip)``
        "warm-up" iterations are performed, but if this took more than
        **duration** seconds, they are accounted towards the measured
        iterations.
    repeat : int
        minimal number of measured iterations to perform.
        Must be positive.
    duration : float
        minimal duration (in seconds) of the measurements.
    """
    def __init__(self, skip=0, repeat=1, duration=0.0):
        self.skip = int(skip)
        self.repeat = int(repeat)
        self.duration = float(duration)

    def time(self, func, *args, **kwargs):
        """
        Repeatedly executes a function at least **repeat** times and for at
        least **duration** seconds (see above), then returns the average time
        per iteration.
        The actual number of measured iterations can be retrieved from
        :py:attr:`Timent.count`.

        Parameters
        ----------
        func : callable
            function to execute
        *args, **kwargs : any, optional
            parameters to pass to **func**

        Returns
        -------
        float
            average function execution time

        Notes
        -----
        The measurements overhead can be estimated by executing ::

            Timent(...).time(lambda: None)

        with a sufficiently large number of iterations (to avoid rounding
        errors due to the finite timer precision).
        In 2018, this overhead was on the order of 100 ns per iteration.
        """
        # Execute "skip" iterations unconditionally
        t0 = timer()
        for i in range(abs(self.skip)):
            func(*args, **kwargs)
        t = timer()
        # if they took longer than "duration" and should be included
        if self.skip < 0 and t - t0 > self.duration:
            # account for them in the "repeat" loop
            start = -self.skip
            if start > self.repeat:
                self.repeat = start
        else:
            # otherwise -- reset the timer and proceed normally
            start = 0
            t0 = timer()

        # Execute "repeat" iterations (maybe accounting for the "skipped")
        for i in range(start, self.repeat):
            func(*args, **kwargs)

        # Continue iterations until "duration" time passes
        for i in itertools.count(self.repeat):
            t = timer()
            if t - t0 > self.duration:
                self.count = i  # save the total number of measured iterations
                break
            func(*args, **kwargs)

        # Return the average time per iteration
        return (t - t0) / self.count


class AbelTiming(object):
    """
    Benchmark performance of different Abel implementations
    (basis generation, forward and inverse transforms, as applicable).

    Parameters
    ----------
    n : int or sequence of int
        array size(s) for the benchmark (assuming 2D square arrays (*n*, *n*))
    select : str or sequence of str
        methods to benchmark. Use ``'all'`` (default) for all available or
        choose any combination of individual methods::

            select=['basex', 'direct_C', 'direct_Python', 'hansenlaw',
                    'linbasex', 'onion_bordas, 'onion_peeling', 'two_point',
                    'three_point']

    repeat : int
        repeat each benchmark at least this number of times to get the average
        values
    t_min : float
        repeat each benchmark for at least this number of seconds to get the
        average values
    t_max : float
        do not benchmark methods at array sizes when this is expected to take
        longer than this number of seconds. Notice that benchmarks for the
        smallest size from **n** are always run and that the estimations can be
        off by a factor of 2 or so.
    verbose : boolean
        determines whether benchmark progress should be reported (to stderr)

    Attributes
    -------
    n : list of int
        array sizes from the parameter **n**, sorted in ascending order
    bs, fabel, iabel : dict of list of float
        benchmark results — dictionaries for

            bs
                basis-set generation
            fabel
                forward Abel transform
            iabel
                inverse Abel transform

        with methods as keys and lists of timings in milliseconds as entries.
        Timings correspond to array sizes in :py:attr:`AbelTiming.n`; for
        skipped benchmarks (see **t_max**) they are ``np.nan``.

    Notes
    -----
    The results can be output in a nice format by simply
    ``print(AbelTiming(...))``.

    Keep in mind that most methods have :math:`O(n^2)` memory and
    :math:`O(n^3)` time complexity, so going from *n* = 501 to *n* = 5001
    would require about 100 times more memory and take about 1000 times longer.
    """
    def __init__(self, n=[301, 501], select='all',
                 repeat=1, t_min=0.1, t_max=np.inf,
                 verbose=True):
        self.n = sorted(_ensure_list(n))
        select = _ensure_list(select)
        self.repeat = repeat
        self.t_max = t_max
        self.verbose = verbose
        # create the timing function
        self._time = Timent(skip=-1, repeat=repeat, duration=t_min).time

        # which methods need half and whole images
        need_half = frozenset([
            'basex',
            'daun',
            'direct_C',
            'direct_Python',
            'hansenlaw',
            'onion_bordas',
            'onion_peeling',
            'two_point',
            'three_point',
        ])
        need_whole = frozenset([
            'linbasex',
            'rbasex',
        ])
        # all available methods (= union of the above sets)
        all_methods = need_half | need_whole
        # remove direct_C, if not supported
        if not direct.cython_ext:
            all_methods = all_methods - frozenset(['direct_C'])

        # Select methods
        if 'all' in select:
            methods = all_methods
        else:
            methods = set()
            for method in select:
                if method not in all_methods:
                    print('Warning: Unsupported method "{}" ignored!'.
                          format(method))
                else:
                    methods.add(method)
        if not methods:
            raise ValueError('At least one valid method must be specified!')

        # dictionary for the results
        self.res = {'bs': {},
                    'forward': {},
                    'inverse': {}}
        # same results as separate dictionaries (aliases to the above)
        self.bs = self.res['bs']
        self.fabel = self.res['forward']
        self.iabel = self.res['inverse']
        # inverse speed for time estimations
        self._pace = {}

        # Loop over all image sizes
        for ni in self.n:
            self.ni = int(ni)
            # image height and half-width
            self.h, self.w = self.ni, self.ni // 2 + 1
            # We transform a rectangular image, since we are making the
            # assumption that we are transforming just the "right side" of
            # a square image.
            # see: https://github.com/PyAbel/PyAbel/issues/207
            self._vprint('n =', self.ni)

            # The following code tries to catch the interruption signal
            # (Ctrl+C) to abort as soon as possible but preserve the available
            # results. Setting a negative time limit makes all remaining
            # benchmarks to skip (calling them is still needed to fill the
            # results with nans).

            # create needed images (half and/or whole)
            if (self.t_max >= 0):  # (do not create while aborting)
                try:
                    if methods & need_half:
                        self.half_image = np.random.randn(self.h, self.w)
                    if methods & need_whole:
                        self.whole_image = np.random.randn(self.h, self.h)
                except (KeyboardInterrupt, MemoryError) as e:
                    self._vprint(repr(e) + ' during image creation!'
                                 ' Skipping the rest...')
                    self.t_max = -1.0
                    # (the images will not be used, so leaving them as is)

            # call benchmark (see below) for each method at this image size
            for method in methods:
                self._vprint(' ', method)
                try:
                    getattr(self, '_time_' + method)()
                except (KeyboardInterrupt, MemoryError) as e:
                    self._vprint('\n' + repr(e) + '! Skipping the rest...')
                    self.t_max = -1.0
                    # rerun this interrupted benchmark to nan-fill its results
                    getattr(self, '_time_' + method)()

            # discard images
            self.half_image = None
            self.whole_image = None
        self._vprint('')

    def _vprint(self, *args, **kwargs):
        """
        Print to stderr, only if verbose=True.
        """
        if self.verbose:
            print(*args, file=sys.stderr, **kwargs)
            sys.stderr.flush()  # (Python 3 buffers stderr. Why?!)

    def _append(self, kind, method, result):
        """
        Store one result, ensuring that the results array exists.
        """
        if method not in self.res[kind]:
            self.res[kind][method] = []
        self.res[kind][method].append(result)

    def _benchmark(self,
                   kind, method,
                   func, *args, **kwargs):
        """
        Run benchmark for the function with arguments and store the result.
        """
        self._append(kind, method,
                     self._time(func, *args, **kwargs) * 1000)  # [s] to [ms]

    def _skip(*param):
        """
        Decorator for benchmarking functions.
        Adds a check whether the estimated execution time would exceed t_max.
        If so, fills the results with np.nan, otherwise executes the
        benchmarking code.

        Parameters are tuples "(kind, method)". Either item can be a list, then
        all combinations of kind(s) and method(s) are implied. Altogether the
        set of these kind–method pairs must be the same as in the "normal"
        execution results.
        """
        # assemble all kind–method pairs
        res_keys = []
        for p in param:
            res_keys += itertools.product(*map(_ensure_list, p))

        def decorator(f):
            method = f.__name__[6:]  # (remove initial "_time_")

            def decorated(self):
                # get the estimated time (use 0 if cannot) and report it
                t_est = self._pace.get(method, 0) * self.ni**3
                self._vprint('    estimated ' +
                             ('{:g} s'.format(_roundsf(t_est, 2)) if t_est
                              else '???'), end='')
                # skip the benchmark if it would take too long
                if t_est > self.t_max:
                    self._vprint(' -- skipped')
                    # fill the results with nan
                    for k, m in res_keys:
                        self._append(k, m, np.nan)
                    return
                else:  # otherwise run the benchmark
                    f(self)
                    # calculate the actual total time and report it
                    t = (sum(self.res[k][m][-1] for k, m in res_keys) *
                         self.repeat) / 1000  # [ms] -> [s]
                    self._vprint(', actually {:.3f} s'.format(t))
                    # save the pace for future estimations
                    self._pace[method] = t / self.ni**3
            return decorated
        return decorator

    # Benchmarking functions for each method.
    # Must be named "_time_method", where "method" is as in "select".
    # Do not take or return anything, but use instance variables:
    # parameters:
    #   self.ni, self.h, self.w -- image size, height, half-width,
    #   self.whole_image, self.half_image -- image (part) to transform
    # results:
    #   self.res[kind][method] = [timings] -- appended for each image size,
    #                                         use np.nan for skipped points
    #     kind = 'bs' (basis), 'forward', 'inverse' -- as applicable
    #     method -- as above, but can also include variants (like in basex)

    @_skip(('bs', 'basex'),
           (['inverse', 'forward'], ['basex', 'basex(var.reg.)']))
    def _time_basex(self):
        # benchmark the basis generation (default parameters)
        def gen_basis():
            basex.cache_cleanup()
            basex.get_bs_cached(self.w, basis_dir=None)
        self._benchmark('bs', 'basex',
                        gen_basis)

        # benchmark all transforms
        for direction in ['inverse', 'forward']:  # (default first)
            # get the transform matrix (default parameters)
            A = basex.get_bs_cached(self.w, basis_dir=None,
                                    direction=direction)
            # benchmark the transform itself
            self._benchmark(direction, 'basex',
                            basex.basex_core_transform,
                            self.half_image, A)

            # benchmark the transform with variable regularization
            def basex_var():
                A = basex.get_bs_cached(self.w, reg=1.0+np.random.random(),
                                        basis_dir=None, direction=direction)
                basex.basex_core_transform(self.half_image, A)
            self._benchmark(direction, 'basex(var.reg.)',
                            basex_var)

            # discard the unneeded transform matrix
            basex.cache_cleanup(direction)

        # discard all caches
        basex.cache_cleanup()

    @_skip(('bs', 'daun'),
           (['inverse', 'forward'], 'daun'),
           ('inverse', 'daun(var.reg.)'))
    def _time_daun(self):
        # benchmark the basis generation (default parameters)
        def gen_basis():
            daun.cache_cleanup()
            daun.get_bs_cached(self.w)
        self._benchmark('bs', 'daun',
                        gen_basis)

        # benchmark all transforms
        for direction in ['inverse', 'forward']:  # (default first)
            # cache the transform matrix
            daun.get_bs_cached(self.w, direction=direction)
            # benchmark the transform itself
            self._benchmark(direction, 'daun',
                            daun.daun_transform,
                            self.half_image, direction=direction,
                            verbose=False)
            if direction == 'inverse':
                # benchmark the transform with variable regularization
                def daun_var():
                    daun.daun_transform(self.half_image,
                                        reg=1.0 + np.random.random(),
                                        verbose=False)
                self._benchmark('inverse', 'daun(var.reg.)',
                                daun_var)
            # discard the unneeded transform matrix
            daun.cache_cleanup(direction)

        # discard all caches
        daun.cache_cleanup()

    @_skip((['inverse', 'forward'], 'direct_C'))
    def _time_direct_C(self):
        for direction in ['inverse', 'forward']:
            self._benchmark(direction, 'direct_C',
                            direct.direct_transform,
                            self.half_image, direction=direction, backend='C')

    @_skip((['inverse', 'forward'], 'direct_Python'))
    def _time_direct_Python(self):
        for direction in ['inverse', 'forward']:
            self._benchmark(direction, 'direct_Python',
                            direct.direct_transform,
                            self.half_image, direction=direction,
                            backend='python')

    @_skip((['inverse', 'forward'], 'hansenlaw'))
    def _time_hansenlaw(self):
        for direction in ['inverse', 'forward']:
            self._benchmark(direction, 'hansenlaw',
                            hansenlaw.hansenlaw_transform,
                            self.half_image, direction=direction)

    @_skip((['bs', 'inverse'], 'linbasex'))
    def _time_linbasex(self):
        # benchmark the basis generation (default parameters)
        def gen_basis():
            linbasex.cache_cleanup()
            linbasex.get_bs_cached(self.h, basis_dir=None)
        self._benchmark('bs', 'linbasex',
                        gen_basis)

        # get the basis (is already cached)
        basis = linbasex.get_bs_cached(self.h, basis_dir=None)
        # benchmark the transform
        self._benchmark('inverse', 'linbasex',
                        linbasex._linbasex_transform_with_basis,
                        self.whole_image, basis)

        # discard all caches
        linbasex.cache_cleanup()

    @_skip(('inverse', 'onion_bordas'))
    def _time_onion_bordas(self):
        self._benchmark('inverse', 'onion_bordas',
                        onion_bordas.onion_bordas_transform,
                        self.half_image)

    # (Generic function for all Dasch methods; not called directly.)
    def _time_dasch(self, method):
        # benchmark the basis generation (default parameters)
        def gen_basis():
            dasch.cache_cleanup()
            dasch.get_bs_cached(method, self.w, basis_dir=None)
        self._benchmark('bs', method,
                        gen_basis)

        # get the transform matrix (is already cached)
        D = dasch.get_bs_cached(method, self.w, basis_dir=None)
        # benchmark the transform
        self._benchmark('inverse', method,
                        dasch.dasch_transform,
                        self.half_image, D)

        # discard all caches
        dasch.cache_cleanup()

    @_skip((['bs', 'inverse'], 'onion_peeling'))
    def _time_onion_peeling(self):
        self._time_dasch('onion_peeling')

    @_skip(('bs', 'rbasex'),
           (['inverse', 'forward'], ['rbasex', 'rbasex(None)']))
    def _time_rbasex(self):
        # benchmark the basis generation (default parameters)
        def gen_basis():
            rbasex.cache_cleanup()
            rbasex.get_bs_cached(self.w - 1)  # Rmax = half-width - 1
        self._benchmark('bs', 'rbasex',
                        gen_basis)

        # benchmark all transforms
        for direction in ['inverse', 'forward']:  # (default first)
            # warm-up run to cache "distributions" (basis is already cached)
            rbasex.rbasex_transform(self.whole_image)
            # benchmark the transform
            self._benchmark(direction, 'rbasex',
                            rbasex.rbasex_transform,
                            self.whole_image)
            # same without output image
            self._benchmark(direction, 'rbasex(None)',
                            rbasex.rbasex_transform,
                            self.whole_image, out=None)

            # discard the unneeded transform matrix
            basex.cache_cleanup(direction)

        # discard all caches
        rbasex.cache_cleanup()

    @_skip((['bs', 'inverse'], 'three_point'))
    def _time_three_point(self):
        self._time_dasch('three_point')

    @_skip((['bs', 'inverse'], 'two_point'))
    def _time_two_point(self):
        self._time_dasch('two_point')

    # (End of benchmarking functions.)

    def __repr__(self):
        import platform

        out = ['PyAbel benchmark run on {}\n'.format(platform.processor()),
               'time in milliseconds']

        # field widths are chosen to accommodate up to:
        #   method = 15 characters
        #   ni = 99999 (would require at least 75 GB RAM)
        #   time = 9999999.9 ms (almost 3 hours)
        # data columns are 9 characters wide and separated by 3 spaces
        TITLE_FORMAT = '=== {} ==='
        HEADER_ROW = 'Method         ' + \
                     ''.join(['   {:>9}'.
                              format('n = {}'.format(ni)) for ni in self.n])
        SEP_ROW = '-' * len(HEADER_ROW)
        ROW_FORMAT = '{:15}' + '   {:9.1f}' * len(self.n)

        def print_benchmark(name, res):
            title = '{:=<{w}}'.format(TITLE_FORMAT.format(name),
                                      w=len(SEP_ROW))
            out = ['\n' + title + '\n']
            out += [HEADER_ROW]
            out += [SEP_ROW]
            for name, row in sorted(res.items()):
                out += [ROW_FORMAT.format(name, *row)]
            return out

        if self.bs:
            out += print_benchmark('Basis generation', self.bs)
            out += ['']
        if self.fabel:
            out += print_benchmark('Forward Abel transform', self.fabel)
            out += ['']
        if self.iabel:
            out += print_benchmark('Inverse Abel transform', self.iabel)

        return '\n'.join(out)


class DistributionsTiming(object):
    """
    Benchmark performance of different VMI distributions implementations.

    Parameters
    ----------
    n : int or sequence of int
        array size(s) for the benchmark (assuming full images to be 2D square
        arrays (*n*, *n*))
    shape : str
        image shape:

        ``'Q'``:
            one quadrant ((*n* + 1)/2, (*n* + 1)/2)
        ``'half'`` (default):
            half image (*n*, (*n* + 1)/2), vertically centered
        ``'full'``:
            full image (*n*, *n*), centered
    rmax : str or sequence of str
        ``'MIN'`` (default) and/or ``'all'``, see **rmax** in
        :class:`abel.tools.vmi.Distributions`
    order : int
        highest order in the angular distributions. Even number ≥ 0.
    weight : str or sequence of str
        weighting to test. Use ``'all'`` for all available or choose any
        combination of individual types::

            weight=['none', 'sin', 'array', 'sin+array']

    method : str or sequence of str
        methods to benchmark. Use ``'all'`` (default) for all available or
        choose any combination of individual methods::

            method=['nearest', 'linear', 'remap']

    repeat : int
        repeat each benchmark at least this number of times to get the average
        values
    t_min : float
        repeat each benchmark for at least this number of seconds to get the
        average values

    Attributes
    -------
    n : list of int
        array sizes from the parameter **n**
    results : dict of dict of dict of list of tuple of float
        benchmark results — multi-level dictionary, in which
        ``results[method][rmax][weight]`` is the list of timings in
        milliseconds corresponding to array sizes in
        :py:attr:`DistributionsTiming.n`. Each timing is a tuple (*t*:sub:`1`,
        *t*:sub:`∞`) with *t*:sub:`1` corresponding to single-image
        (non-cached) performance, and *t*:sub:`∞` corresponding to batch
        (cached) performance.

    Notes
    -----
    The results can be output in a nice format by simply
    ``print(DistributionsTiming(...))``.
    """
    def __init__(self, n=[301, 501], shape='half', rmax='MIN', order=2,
                 weight=['none', 'sin', 'sin+array'], method='all',
                 repeat=1, t_min=0.1):
        self.n = _ensure_list(n)

        if shape == 'Q':
            origin = 'll'
            self.shape = 'One quadrant'
        elif shape == 'half':
            origin = 'cl'
            self.shape = 'Half image'
        elif shape == 'full':
            origin = 'cc'
            self.shape = 'Full image'
        else:
            raise ValueError('Incorrect shape "{}"'.format(shape))

        self.rmaxs = rmaxs = _ensure_list(rmax)

        self.order = order

        weights = _ensure_list(weight)
        if 'all' in weights:
            weights = ['none', 'sin', 'array', 'sin+array']
        self.weights = weights

        methods = _ensure_list(method)
        if 'all' in methods:
            methods = ['nearest', 'linear', 'remap']
        self.methods = methods

        # create the timing function
        time = Timent(skip=-1, repeat=repeat, duration=t_min).time

        # dictionary for the results
        self.results = {m: {r: {w: [] for w in weights}
                                      for r in rmaxs}
                                      for m in methods}

        from abel.tools.vmi import Ibeta, Distributions
        # make sure that everything is loaded
        # (otherwise the 1st timing call is very slow)
        Ibeta(np.array([[0]]))

        # Loop over all image sizes
        for ni in self.n:
            ni = int(ni)

            # create image and weight array
            rows = (ni + 1) // 2 if shape == 'Q' else ni
            cols = (ni + 1) // 2 if shape in ['Q', 'half'] else ni
            IM = np.random.randn(rows, cols)
            warray = np.random.randn(rows, cols)

            # benchmark each combination of parameters
            for method in methods:
                for rmax in rmaxs:
                    for weight in weights:
                        if weight == 'none':
                            w = {'use_sin': False, 'weights': None}
                        elif weight == 'sin':
                            w = {'use_sin': True, 'weights': None}
                        elif weight == 'array':
                            w = {'use_sin': False, 'weights': warray}
                        elif weight == 'sin+array':
                            w = {'use_sin': True, 'weights': warray}
                        else:
                            raise ValueError('Incorrect weight "{}"'.
                                             format(weight))
                        # single-image
                        t1 = time(Ibeta,
                                  IM, origin, rmax, order, method=method, **w)
                        # cached
                        distr = Distributions(origin, rmax, order,
                                              method=method, **w)
                        distr(IM)  # trigger precalculations

                        def distrIMIbeta(IM):
                            return distr(IM).Ibeta()
                        tn = time(distrIMIbeta,
                                  IM)
                        # save results
                        self.results[method][rmax][weight].append((t1 * 1000,
                                                                   tn * 1000))

    def __repr__(self):
        import platform

        out = ['PyAbel benchmark run on {}\n'.format(platform.processor()),
               'order = {}, time in milliseconds'.format(self.order)]

        # field widths are chosen to accommodate up to:
        #   rmax + weight = 3 leading spaces + 14 characters
        #   ni = 99999
        #   time = 9999999.9 ms (almost 3 hours)
        # data columns are 9 characters wide and separated by 3 spaces
        TITLE_FORMAT = '=== ' + self.shape + ', {} ==='
        HEADER_ROW = 'Method           ' + \
                     ''.join(['   {:>9}'.
                              format('n = {}'.format(ni)) for ni in self.n])
        SEP_ROW = '-' * len(HEADER_ROW)
        ROW_FORMAT = '   {}, {:9}' + '   {:9.1f}' * len(self.n)

        def print_benchmark(mode):
            title = '{:=<{w}}'.format(TITLE_FORMAT.format(mode),
                                      w=len(SEP_ROW))
            out = ['\n' + title + '\n']
            out += [HEADER_ROW]
            out += [SEP_ROW]
            for method in self.methods:
                out += [method]
                resm = self.results[method]
                for rmax in self.rmaxs:
                    resr = resm[rmax]
                    for weight in self.weights:
                        res = resr[weight]
                        t = list(zip(*res))[0 if mode == 'single' else 1]
                        out += [ROW_FORMAT.format(rmax, str(weight), *t)]
            return out

        out += print_benchmark('single')
        out += ['']
        out += print_benchmark('cached')

        return '\n'.join(out)


def is_symmetric(arr, i_sym=True, j_sym=True):
    """
    Takes in an array of shape (n, m) and check if it is symmetric

    Parameters
    ----------
    arr : 1D or 2D array
    i_sym : array
        symmetric with respect to the 1st axis
    j_sym : array
        symmetric with respect to the 2nd axis

    Returns
    -------
    a binary array with the symmetry condition for the corresponding quadrants.
    The globa

    Notes
    -----
    If both **i_sym** = ``True`` and **j_sym** = ``True``, the input array is
    checked for polar symmetry.

    See `issue #34 comment
    <https://github.com/PyAbel/PyAbel/issues/34#issuecomment-160344809>`__
    for the defintion of a center of the image.
    """

    Q0, Q1, Q2, Q3 = tools.symmetry.get_image_quadrants(
                                                 arr, reorient=False)

    if i_sym and not j_sym:
        valid_flag = [np.allclose(np.fliplr(Q1), Q0),
                      np.allclose(np.fliplr(Q2), Q3)]
    elif not i_sym and j_sym:
        valid_flag = [np.allclose(np.flipud(Q1), Q2),
                      np.allclose(np.flipud(Q0), Q3)]
    elif i_sym and j_sym:
        valid_flag = [np.allclose(np.flipud(np.fliplr(Q1)), Q3),
                      np.allclose(np.flipud(np.fliplr(Q0)), Q2)]
    else:
        raise ValueError('Checking for symmetry with both i_sym=False \
                          and j_sym=False does not make sense!')

    return np.array(valid_flag)


def absolute_ratio_benchmark(analytical, recon, kind='inverse'):
    """
    Check the absolute ratio between an analytical function and the result
    of a inverse Abel reconstruction.

    Parameters
    ----------
    analytical : one of the classes from analytical, initialized

    recon : 1D ndarray
        a reconstruction (i.e. inverse abel)
        given by some PyAbel implementation
    """
    mask = analytical.mask_valid

    if kind == 'inverse':
        func = analytical.func
    elif kind == 'forward':
        func = analytical.abel

    err = func[mask]/recon[mask]
    return err



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import abel
import os
from glob import glob
import scipy
from scipy.special import eval_legendre
from scipy import ndimage

###############################################################################
# linbasex - inversion procedure based on 1-dimensional projections of
#            velocity-map images
#
# As described in:
#   Gerber, Thomas, Yuzhu Liu, Gregor Knopp, Patrick Hemberger, Andras Bodi,
#   Peter Radi, and Yaroslav Sych,
#     “Charged Particle Velocity Map Image Reconstruction with One-Dimensional
#      Projections of Spherical Functions.”
#     Review of Scientific Instruments 84, no. 3 (March 1, 2013):
#                                      033101–033101 – 10.
#     doi:10.1063/1.4793404.
#
# 2016-04- Thomas Gerber and Daniel Hickstein - theory and code updates
# 2016-04- Stephen Gibson core code extracted from the supplied jupyter
#          notebook (see #167: https://github.com/PyAbel/PyAbel/issues/167)
#
###############################################################################

_linbasex_parameter_docstring = \
    r"""Inverse Abel transform using 1d projections of images.

    Th. Gerber, Yu. Liu, G. Knopp, P. Hemberger, A. Bodi, P. Radi, Ya. Sych,
    "Charged particle velocity map image reconstruction with one-dimensional
    projections of spherical functions",
    `Rev. Sci. Instrum. 84, 033101 (2013)
    <https://doi.org/10.1063/1.4793404>`__.

    ``linbasex`` models the image using a sum of Legendre polynomials at each
    radial pixel, As such, it should only be applied to situations that can
    be adequately represented by Legendre polynomials, i.e., images that
    feature spherical-like structures.  The reconstructed 3D object is
    obtained by adding all the contributions, from which slices are derived.


    Parameters
    ----------
    IM : numpy 2D array
        image data must be square shape of odd size
    proj_angles : list
        projection angles, in radians (default :math:`[0, \pi/2]`)
        e.g. :math:`[0, \pi/2]` or :math:`[0, 0.955, \pi/2]` or :math:`[0, \pi/4, \pi/2, 3\pi/4]`
    legendre_orders : list
        orders of Legendre polynomials to be used as the expansion

        * even polynomials [0, 2, ...] gerade
        * odd polynomials [1, 3, ...] ungerade
        * all orders [0, 1, 2, ...].

        In a single photon experiment there are only anisotropies up to
        second order. The interaction of 4 photons (four wave mixing) yields
        anisotropies up to order 8.
    radial_step : int
        number of pixels per Newton sphere (default 1)
    smoothing: float
        convolve Beta array with a Gaussian function of 1/e 1/2 width `smoothing`.
    rcond : float
        (default 0.0005) scipy.linalg.lstsq fit conditioning value.
        set rcond to zero to switch conditioning off.
        Note: In the presence of noise the equation system may be ill posed.
        Increasing rcond smoothes the result, lowering it beyond a minimum
        renders the solution unstable. Tweak rcond to get a "reasonable"
        solution with acceptable resolution.
    clip : int
        clip first vectors (smallest Newton spheres) to avoid singularities
        (default 0)
    norm_range : tuple
        (low, high)
        normalization of Newton spheres, maximum in range Beta[0, low:high].
        Note: Beta[0, i] the total number of counts integrated over sphere i,
        becomes 1.
    threshold : float
        threshold for normalization of higher order Newton spheres (default 0.2)
        Set all Beta[j], j>=1 to zero if the associated Beta[0] is smaller
        than threshold.
    return_Beta : bool
        return the Beta array of Newton spheres, as the tuple: radial-grid, Beta
        for the case :attr:`legendre_orders=[0, 2]`

            Beta[0] vs radius -> speed distribution

            Beta[2] vs radius -> anisotropy of each Newton sphere

        see 'Returns'.
    direction : str
        "inverse" - only option for this method.
        Abel transform direction.
    dr : None
        dummy variable for call compatibility with the other methods
    verbose : bool
        print information about processing (normally used for debugging)


    Returns
    -------
    inv_IM : numpy 2D array
       inverse Abel transformed image

    radial, Beta, projections : tuple
       (if :attr:`return_Beta=True`)

       contributions of each spherical harmonic :math:`Y_{i0}` to the 3D
       distribution contain all the information one can get from an experiment.
       For the case :attr:`legendre_orders=[0, 2]`:

           Beta[0] vs radius -> speed distribution

           Beta[1] vs radius -> anisotropy of each Newton sphere.

       projections : are the radial projection profiles at angles `proj_angles`

    """

# cache basis
_basis = None
_los = None   # legendre_orders string
_pas = None   # proj_angles string
_radial_step = None
_clip = None


def linbasex_transform(IM, basis_dir=None, proj_angles=[0, np.pi/2],
                       legendre_orders=[0, 2], radial_step=1, smoothing=0,
                       rcond=0.0005, threshold=0.2, return_Beta=False, clip=0,
                       norm_range=(0, -1), direction="inverse", verbose=False,
                       dr=None):
    """
    Wrapper function for linbasex to process supplied quadrant-image as a
    full-image.

    PyAbel transform functions operate on the right side of an image.
    Here we follow the `basex` technique of duplicating the right side to
    the left re-forming the whole image.

    """
    IM = np.atleast_2d(IM)

    quad_rows, quad_cols = IM.shape
    full_image = abel.tools.symmetry.put_image_quadrants((IM, IM, IM, IM),
                 original_image_shape=(quad_rows*2-1, quad_cols*2-1))

    # inverse Abel transform
    recon, radial, Beta, QLz = linbasex_transform_full(full_image,
                               basis_dir=basis_dir, proj_angles=proj_angles,
                               legendre_orders=legendre_orders,
                               radial_step=radial_step, smoothing=smoothing,
                               threshold=threshold, clip=clip,
                               norm_range=norm_range,
                               verbose=verbose)

    # unpack right-side
    inv_IM = abel.tools.symmetry.get_image_quadrants(recon)[0]

    if return_Beta:
        return inv_IM, radial, Beta, QLz
    else:
        return inv_IM


def linbasex_transform_full(IM, basis_dir=None, proj_angles=[0, np.pi/2],
                            legendre_orders=[0, 2],
                            radial_step=1, smoothing=0,
                            rcond=0.0005, threshold=0.2, clip=0,
                            return_Beta=False, norm_range=(0, -1),
                            direction="inverse", verbose=False):
    """interface function that fetches/calculates the Basis and
       then evaluates the linbasex inverse Abel transform for the image.

    """

    IM = np.atleast_2d(IM)

    rows, cols = IM.shape

    if cols % 2 == 0:
        raise ValueError('image width ({}) must be odd and equal to the height'
                         .format(cols))

    if rows != cols:
        raise ValueError('image has shape ({}, {}), '.format(rows, cols) +
                         'must be square for a "linbasex" transform')

    # generate basis or read from file if available
    _basis = get_bs_cached(cols, basis_dir=basis_dir, proj_angles=proj_angles,
                  legendre_orders=legendre_orders, radial_step=radial_step,
                  clip=clip, verbose=verbose)

    return _linbasex_transform_with_basis(IM, _basis, proj_angles=proj_angles,
                                    legendre_orders=legendre_orders,
                                    radial_step=radial_step,
                                    rcond=rcond, smoothing=smoothing,
                                    threshold=threshold, clip=clip,
                                    norm_range=norm_range)


def _linbasex_transform_with_basis(IM, Basis, proj_angles=[0, np.pi/2],
                                   legendre_orders=[0, 2], radial_step=1,
                                   rcond=0.0005, smoothing=0, threshold=0.2,
                                   clip=0, norm_range=(0, -1)):
    """linbasex inverse Abel transform evaluated with supplied basis set Basis.

    """

    IM = np.atleast_2d(IM)

    rows, cols = IM.shape

    # Number of used polynoms
    pol = len(legendre_orders)

    # How many projections
    proj = len(proj_angles)

    QLz = np.zeros((proj, cols))  # array for projections.

    # Rotate and project VMI-image for each angle (as many as projections)
    # if proj_angles == [0, np.pi/2]:
    #     # If coordinates of the detector coincide with the projection
    #     # directions unnecessary rotations are avoided
    #     # i.e. proj_angles=[0, np.pi/2] degrees
    #     QLz[0] = np.sum(IM, axis=1)
    #     QLz[1] = np.sum(IM, axis=0)
    # else:
    for i in range(proj):
        Rot_IM = scipy.ndimage.interpolation.rotate(IM,
                       proj_angles[i]*180/np.pi, axes=(1, 0), reshape=False)
        QLz[i, :] = np.sum(Rot_IM, axis=1)

    # arrange all projections for input into "lstsq"
    bb = np.concatenate(QLz, axis=0)

    Beta = _beta_solve(Basis, bb, pol, rcond=rcond)

    inv_IM, Beta_convol = _Slices(Beta, legendre_orders, smoothing=smoothing)

    # normalize
    Beta = _single_Beta_norm(Beta_convol, threshold=threshold,
                             norm_range=norm_range)

    radial = np.linspace(clip, cols//2, len(Beta[0]))

    # Fix Me! Issue #202 the correct scaling factor for inv_IM intensity?
    return inv_IM, radial, Beta, QLz


linbasex_transform_full.__doc__ = _linbasex_parameter_docstring


def _beta_solve(Basis, bb, pol, rcond=0.0005):
    # set rcond to zero to switch conditioning off

    # array for solutions. len(Basis[0,:])//pol is an integer.
    Beta = np.zeros((pol, len(Basis[0])//pol))

    # solve equation
    Sol = np.linalg.lstsq(Basis, bb, rcond)

    # arrange solutions into subarrays for each β.
    Beta = Sol[0].reshape((pol, len(Sol[0])//pol))

    return Beta


def _SL(i, x, y, Beta_convol, index, legendre_orders):
    """Calculates interpolated β(r), where r= radius"""
    r = np.sqrt(x**2 + y**2 + 0.1)  # + 0.1 to avoid divison by zero.

    # normalize: divison by circumference.
    # @stggh 2/r to correctly normalize intensity cf O2 PES
    BB = np.interp(r, index, Beta_convol[i, :], left=0)/(4*np.pi*r*r)

    return BB*eval_legendre(legendre_orders[i], x/r)


def _Slices(Beta, legendre_orders, smoothing=0):
    """Convolve Beta with a Gaussian function of 1/e width smoothing.

    """

    pol = len(legendre_orders)
    NP = len(Beta[0])  # number of points in 3_d plot.
    index = range(NP)

    Beta_convol = np.zeros((pol, NP))
    Slice_3D = np.zeros((pol, 2*NP, 2*NP))

    # Convolve Beta's with smoothing function
    if smoothing > 0:
        # smoothing function
        Basis_s = np.fromfunction(lambda i: np.exp(-(i - (NP)/2)**2 /
                                  (2*smoothing**2))/(smoothing*2.5), (NP,))
        for i in range(pol):
            Beta_convol[i] = np.convolve(Basis_s, Beta[i], mode='same')
    else:
        Beta_convol = Beta

    # Calculate ordered slices:
    for i in range(pol):
        Slice_3D[i] = np.fromfunction(lambda k, l: _SL(i, (k-NP), (l-NP),
                       Beta_convol, index, legendre_orders), (2*NP, 2*NP))
    # Sum ordered slices up
    Slice = np.sum(Slice_3D, axis=0)

    return Slice, Beta_convol


def int_beta(Beta, radial_step=1, threshold=0.1, regions=None):
    """Integrate beta over a range of Newton spheres.

    Parameters
    ----------
    Beta : numpy array
        Newton spheres
    radial_step : int
        number of pixels per Newton sphere (default 1)
    threshold : float
        threshold for normalisation of higher orders, 0.0 ... 1.0.
    regions : list of tuple radial ranges
        [(min0, max0), (min1, max1), ...]

    Returns
    -------
    Beta_in : numpy array
        integrated normalized Beta array [Newton sphere, region]

    """
    pol = Beta.shape[0]
    # Define new array for normalized beta's, independent of Beat_norm
    Beta_n = np.zeros(Beta.shape)

    # Normalized to Newton sphere with maximal counts.
    max_counts = max(Beta[0, :])

    Beta_n[0] = Beta[0]/max_counts
    for i in range(1, pol):
        Beta_n[i] = np.where(Beta[0]/max_counts > threshold, Beta[i]/Beta[0],
                             0)

    Beta_int = np.zeros((pol, len(regions)))   # arrays for results

    for j, reg in enumerate(regions):
        for i in range(pol):
            Beta_int[i, j] = np.sum(Beta_n[i, range(*reg)])/(reg[1]-reg[0])

    return Beta_int


def _single_Beta_norm(Beta, threshold=0.2, norm_range=(0, -1)):
    """Normalize Newton spheres.

    Parameters
    ----------
    Beta : numpy array
        Newton spheres
    threshold : float
        choose only Beta's for which Beta0 is greater than the maximal Beta0
        times threshold in the chosen range
        Set all βi, i>=1 to zero if the associated β0 is smaller than threshold

    norm_range : tuple (int, int)
        (low, high)
        normalize to Newton sphere with maximum counts in chosen range.
        Beta[0, low:high]

    Return
    ------
    Beta : numpy array
        normalized Beta array

    """
    pol = Beta.shape[0]

    Beta_norm = np.zeros_like(Beta)
    # Normalized to Newton sphere with maximum counts in chosen range.
    max_counts = Beta[0, norm_range[0]:norm_range[1]].max()

    Beta_norm[0] = Beta[0]/max_counts
    for i in range(1, pol):
        Beta_norm[i] = np.where(Beta[0]/max_counts > threshold,
                                Beta[i]/Beta[0], 0)

    return Beta_norm


def _bas(ord, angle, COS, TRI):
    """Define Basis vectors for a given polynomial order "order" and a
       given projection angle "angle".

    """

    basis_vec = scipy.special.eval_legendre(ord, angle) *\
                scipy.special.eval_legendre(ord, COS) * TRI
    return basis_vec


def _bs_linbasex(cols, proj_angles=[0, np.pi/2], legendre_orders=[0, 2],
                 radial_step=1, clip=0):

    pol = len(legendre_orders)
    proj = len(proj_angles)

    # Calculation of Base vectors
    # Define triangular matrix containing columns :math:`x/y` (representing :math:`\cos(\theta))`.
    n = cols//2 + cols % 2
    Index = np.indices((n, n))
    Index[:, 0, 0] = 1
    cos = Index[0]*np.tri(n, n, k=0)[::-1, ::-1]/np.diag(Index[0])

    # Concatenate to "bi"-triangular matrix
    COS = np.concatenate((-cos[::-1, :], cos[1:, :]), axis=0)
    TRI = np.concatenate((np.tri(n, n, k=0,)[:-1, ::-1],
                          np.tri(n, n, k=0,)[::-1, ::-1]), axis=0)

    # radial_step: use only each radial_step other vector.
    # Keep the base vector with full span
    COS = COS[:, ::-radial_step]
    TRI = TRI[:, ::-radial_step]

    COS = COS[:, ::-1]  # rearrange base vectors again in ascending order
    TRI = TRI[:, ::-1]

    if clip > 0:
        # clip first vectors (smallest Newton spheres) to avoid singularities
        COS = COS[:, clip:]
        # It is difficult to trace the effect on the SVD solver used below.
        TRI = TRI[:, clip:]  # usually no clipping works fine.

    # Calculate base vectors for each projection and each order.
    B = np.zeros((pol, proj, len(COS[:, 0]), len(COS[0, :])))

    Norm = np.sum(_bas(0, 1, COS, TRI), axis=0)  # calculate normalization
    cos_an = np.cos(proj_angles)  # angles in radians

    for p in range(pol):
        for u in range(proj):
            B[p, u] = _bas(legendre_orders[p], cos_an[u], COS, TRI)/Norm

    # concatenate vectors to one matrix of bases
    Bpol = np.concatenate((B), axis=2)
    Basis = np.concatenate((Bpol), axis=0)

    return Basis


def get_bs_cached(cols, basis_dir=None, legendre_orders=[0, 2],
                  proj_angles=[0, np.pi/2],
                  radial_step=1, clip=0, verbose=False):
    """load basis set from disk, generate and store if not available.

    Checks whether file:
    ``linbasex_basis_{cols}_{legendre_orders}_{proj_angles}_{radial_step}_{clip}*.npy`` is present in `basis_dir`

    Either, read basis array or generate basis, saving it to the file.


    Parameters
    ----------
    cols : int
        width of image

    basis_dir : str or None
        path to the directory for saving / loading the basis. Use ``''`` for
        the default directory. If ``None``, the basis set will not be loaded
        from or saved to disk.

    legendre_orders : list
        default [0, 2] = 0 order and 2nd order polynomials

    proj_angles : list
        default [0, np.pi/2] in radians

    radial_step : int
        pixel grid size, default 1

    clip : int
        image edge clipping, default 0 pixels

    verbose: boolean
        print information for debugging

    Returns
    -------
    D : tuple (B, Bpol)
       of ndarrays B (pol, proj, cols, cols) Bpol (pol, proj)

    file.npy: file
       saves basis to file name ``linbasex_basis_{cols}_{legendre_orders}_{proj_angles}_{radial_step}_{clip}.npy``

    """

    # cached basis
    global _basis, _los, _pas, _radial_step, _clip

    # legendre_orders string
    los = ''.join(map(str, legendre_orders))
    # convert to % of pi
    proj_angles_fractpi = np.array(proj_angles)*100/np.pi
    # projection angles string
    pas = ''.join(map(str, proj_angles_fractpi.astype(int)))

    if _basis is not None:
        # check basis array sizes, warning may not be unique
        if _basis.shape == (2*cols, cols+1):
            if _los == los and _pas == pas and _radial_step == radial_step and\
               _clip == clip:
                if verbose:
                    print('Using memory cached basis')
                return _basis

    # Fix Me! not a simple unique naming mechanism
    basis_name = "linbasex_basis_{}_{}_{}_{}_{}.npy".format(cols, los, pas,
                                                            radial_step, clip)

    _los = los
    _pas = pas
    _radial_step = radial_step
    _clip = clip
    if basis_dir == '':
        basis_dir = abel.transform.get_basis_dir(make=True)
    if basis_dir is not None:
        path_to_basis_file = os.path.join(basis_dir, basis_name)
        if os.path.exists(path_to_basis_file):
            if verbose:
                print('loading {} ...'.format(path_to_basis_file))
            _basis = np.load(path_to_basis_file)
            return _basis

    if verbose:
        print("A suitable basis for linbasex was not found.\n"
              "A new basis will be generated.")

    _basis = _bs_linbasex(cols, proj_angles=proj_angles,
                     legendre_orders=legendre_orders, radial_step=radial_step,
                     clip=clip)

    if basis_dir is not None:
        path_to_basis_file = os.path.join(basis_dir, basis_name)
        np.save(path_to_basis_file, _basis)
        if verbose:
            print("linbasex basis saved for later use to {}"
                  .format(path_to_basis_file))

    return _basis


def cache_cleanup():
    """
    Utility function.

    Frees the memory caches created by ``get_bs_cached()``.
    This is usually pointless, but might be required after working
    with very large images, if more RAM is needed for further tasks.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """

    global _basis, _los, _pas, _radial_step, _clip

    _basis = None
    _los = None
    _pas = None
    _radial_step = None
    _clip = None


def basis_dir_cleanup(basis_dir=''):
    """
    Utility function.

    Deletes basis sets saved on disk.

    Parameters
    ----------
    basis_dir : str or None
        relative or absolute path to the directory with saved basis sets. Use
        ``''`` for the default directory. ``None`` does nothing.

    Returns
    -------
    None
    """
    if basis_dir == '':
        basis_dir = abel.transform.get_basis_dir(make=False)

    if basis_dir is None:
        return

    files = glob(os.path.join(basis_dir, 'linbasex_basis_*.npy'))
    for fname in files:
        os.remove(fname)


linbasex_transform.__doc__ += _linbasex_parameter_docstring



# -*- coding: utf-8 -*-
from __future__ import division
import numpy as np
from scipy.ndimage.interpolation import map_coordinates
from scipy.interpolate import UnivariateSpline, splrep, splev
from scipy.optimize import leastsq

import abel

from abel import _deprecated, _deprecate

#########################################################################
# circularize.py
#
# Image circularization by following peak intensity vs angle
# see https://github.com/PyAbel/PyAbel/issues/186 for discussion
# and https://github.com/PyAbel/PyAbel/pull/195
#
# Steve Gibson and Dan Hickstein - ideas/code
# Jason Gascooke - ideas
#
# February 2017
#########################################################################


def circularize_image(IM, method="lsq", origin=None, radial_range=None,
                      dr=0.5, dt=0.5, smooth=_deprecated, ref_angle=None,
                      inverse=False, return_correction=False, tol=0,
                      center=_deprecated):
    r"""
    Corrects image distortion on the basis that the structure should be
    circular.

    This is a simplified radial scaling version of the algorithm described in
    J. R. Gascooke, S. T. Gibson, W. D. Lawrance,
    "A 'circularisation' method to repair deformations and determine the centre
    of velocity map images",
    `J. Chem. Phys. 147, 013924 (2017)
    <https://dx.doi.org/10.1063/1.4981024>`_.

    This function is especially useful for correcting the image obtained with
    a velocity-map-imaging spectrometer, in the case where there is distortion
    of the Newton sphere (ring) structure due to an imperfect electrostatic
    lens or stray electromagnetic fields. The correction allows the
    highest-resolution 1D photoelectron distribution to be extracted.

    The algorithm splits the image into "slices" at many different angles
    (set by **dt**) and compares the radial intensity profile of adjacent slices.
    A scaling factor is found which aligns each slice profile with the previous
    slice. The image is then corrected using a spline function that smoothly
    connects the discrete scaling factors as a continuous function of angle.

    This circularization algorithm should only be applied to a well-centered
    image, otherwise use the **origin** keyword (described below) to
    center it.


    Parameters
    ----------
    IM : numpy 2D array
        Image to be circularized.

    method : str
        Method used to determine the radial correction factor to align slice
        profiles:

        ``argmax``
            compare intensity-profile.argmax() of each radial slice.
            This method is quick and reliable, but it assumes that
            the radial intensity profile has an obvious maximum.
            The positioning is limited to the nearest pixel.

        ``lsq``
            minimize the difference between a slice intensity-profile
            with its adjacent slice.
            This method is slower and may fail to converge, but it
            may be applied to images with any (circular) structure.
            It aligns the slices with sub-pixel precision.

    origin : float tuple, str or None
        Pre-center image using :func:`abel.tools.center.center_image`.
        May be an explicit (row, column) tuple or a method name: ``'com'``,
        ``'convolution'``, ``'gaussian;``, ``'image_center'``, ``'slice'``.
        ``None`` (default) assumes that the image is already centered.

    radial_range : tuple or None
        Limit slice comparison to the radial range tuple (rmin, rmax), in
        pixels, from the image origin. Use to determine the distortion
        correction associated with particular peaks. It is recommended to
        select a region of your image where the signal-to-noise ratio is
        highest, with sharp persistent (in angle) features.

    dr : float
        Radial grid size for the polar coordinate image, default = 0.5 pixel.
        This is passed to :func:`abel.tools.polar.reproject_image_into_polar`.

        Small values may improve the distortion correction, which is often of
        sub-pixel dimensions, at the cost of reduced signal to noise for the
        slice intensity profile. As a general rule, `dr` should be
        significantly smaller than the radial "feature size" in the image.

    dt : float
        Angular grid size. This sets the number of radial slices, given by
        :math:`2\pi/dt`. Default = 0.1, ~ 63 slices. More slices, using
        smaller `dt`, may provide a more detailed angular variation of the
        correction, at the cost of greater signal to noise in the correction
        function.

        Also passed to :func:`abel.tools.polar.reproject_image_into_polar`.

    smooth : float
        Deprecated, use **tol** instead. The relationship is
        **smooth** = `N`\ :sub:`angles` × **tol**\ :sup:`2`,
        where `N`\ :sub:`angles` is the number of slices (see **dt**).

    ref_angle : None or float
        Reference angle for which radial coordinate is unchanged.
        Angle varies between :math:`-\pi` and :math:`\pi`, with zero angle
        vertical.

        ``None`` uses :func:`numpy.mean` of the radial correction function,
        which attempts to maintain the same average radial scaling. This
        approximation is likely valid, unless you know for certain that a
        specific angle of your image corresponds to an undistorted image.

    inverse : bool
        Apply an inverse Abel transform to the `polar`-coordinate image, to
        remove the background intensity. This may improve the signal-to-noise
        ratio, allowing the weaker intensity featured to be followed in angle.

        Note that this step is only for the purposes of allowing the algorithm
        to better follow peaks in the image. It does not affect the final
        image that is returned, except for (hopefully) slightly improving the
        precision of the distortion correction.

    return_correction : bool
        Additional outputs, as describe below.

    tol : float
        Root-mean-square (RMS) fitting tolerance for the spline function. At
        the default zero value, the spline interpolates between the discrete
        scaling factors. At larger values, a smoother spline is found such that
        its RMS deviation from the discrete scaling factors does not exceed
        this number. For example, ``tol=0.01`` means 1% RMS tolerance for the
        radial scaling correction. At very large tolerances, the spline
        degenerates to a constant, the average of the discrete scaling factors.

        Typically, **tol** may remain zero (use interpolation), but noisy data
        may require some smoothing, since the found discrete scaling factors
        can have noticeable errors. To examine the relative scaling factors and
        how well they are represented by the spline function, use the option
        ``return_correction=True``.

    Returns
    -------
    IMcirc : numpy 2D array
        Circularized version of the input image, same size as input.

    The following values are returned if ``return_correction=True``:

    angles : numpy 1D array
        Mid-point angle (radians) of each image slice.

    radial_correction : numpy 1D array
        Radial correction scale factor at each angular slice.

    radial_correction_function : function(numpy 1D array)
        Function that may be used to evaluate the radial correction at any
        angle.

    """
    if center is not _deprecated:
        _deprecate('abel.tools.circularize.circularize_image() '
                   'argument "center" is deprecated, use "origin" instead.')
        origin = center

    if origin is not None:
        # convenience function for the case image is not centered
        IM = abel.tools.center.center_image(IM, method=origin)

    # map image into polar coordinates - much easier to slice
    # cartesian (Y, X) -> polar (Radius, Theta)
    polarIM, radial_coord, angle_coord = \
        abel.tools.polar.reproject_image_into_polar(IM, dr=dr, dt=dt)

    if inverse:
        # pseudo inverse Abel transform of the polar image, removes background
        # to enhance transition peaks
        polarIM = abel.dasch.two_point_transform(polarIM.T).T

    # more convenient 1-D coordinate arrays
    angles = angle_coord[0]  # angle coordinate
    radial = radial_coord[:, 0]  # radial coordinate

    # limit radial range of polar image, if selected
    if radial_range is not None:
        subr = np.logical_and(radial > radial_range[0],
                              radial < radial_range[1])
        polarIM = polarIM[subr]
        radial = radial[subr]

    # evaluate radial correction factor that aligns each angular slice
    radcorr = correction(polarIM.T, angles, radial, method=method)

    if smooth is not _deprecated:
        _deprecate('abel.tools.circularize.circularize_image() '
                   'argument "smooth" is deprecated, use "tol" instead.')
    else:
        smooth = len(angles) * tol**2

    # periodic spline radial correction vs angle
    spl = splrep(np.append(angles, angles[0] + 2 * np.pi),
                 np.append(radcorr, radcorr[0]), s=smooth, per=True)

    def radial_correction_function(angle):
        return splev(angle, spl)

    # apply the correction
    IMcirc = circularize(IM, radial_correction_function, ref_angle=ref_angle)

    if return_correction:
        return IMcirc, angles, radcorr, radial_correction_function
    else:
        return IMcirc


def circularize(IM, radial_correction_function, ref_angle=None):
    """
    Remap image from its distorted grid to the true cartesian grid.

    Parameters
    ----------
    IM : numpy 2D array
        Original image

    radial_correction_function : function(numpy 1D array)
        A function returning the radial correction for a given angle. It
        should accept a numpy 1D array of angles.

    """
    # cartesian coordinate system
    Y, X = np.indices(IM.shape)

    row, col = IM.shape
    origin = (col // 2, row // 2)  # odd image

    # coordinates relative to center
    X -= origin[0]
    Y = origin[1] - Y   # negative values below the axis
    theta = np.arctan2(X, Y)  # referenced to vertical direction

    # radial scale factor at angle = ref_angle
    if ref_angle is None:
        factor = np.mean(radial_correction_function(theta))
    else:
        factor = radial_correction_function(ref_angle)

    # radial correction
    Xactual = X * factor / radial_correction_function(theta)
    Yactual = Y * factor / radial_correction_function(theta)

    # @DanHickstein magic
    # https://github.com/PyAbel/PyAbel/issues/186#issuecomment-275471271
    IMcirc = map_coordinates(IM, (origin[1] - Yactual, Xactual + origin[0]))

    return IMcirc


def _residual(param, radial, profile, previous):
    """ `scipy.optimize.leastsq` residuals function.

        Evaluate the difference between a radial-scaled intensity profile
        and its adjacent "previous" angular slice.

    """

    radial_scaling, amplitude = param[0], param[1]

    newradial = radial * radial_scaling
    spline_prof = UnivariateSpline(newradial, profile, s=0, ext=3)
    newprof = spline_prof(radial) * amplitude

    # residual cf adjacent slice profile
    return newprof - previous


def correction(polarIMTrans, angles, radial, method):
    r"""
    Determines a radial correction factors that align an angular slice
    radial intensity profile with its adjacent (previous) slice profile.


    Parameters
    ----------
    polarIMTrans : numpy 2D array
        Polar coordinate image, transposed :math:`(\theta, r)` so that each
        row is a single angle.

    angles : numpy 1D array
        Angle coordinates for one row of `polarIMTrans`.

    radial : numpy 1D array
        Radial coordinates for one column of `polarIMTrans`.

    method : str
        ``argmax``
            radial correction factor from position of maximum intensity.

        ``lsq``
            least-squares determine a radial correction factor that will align
            a radial intensity profile with the previous, adjacent slice.
    """

    if method == "argmax":
        # follow position of intensity maximum
        pkpos = []

        for ang, aslice in zip(angles, polarIMTrans):
            profile = aslice
            pkpos.append(profile.argmax())  # store index of peak position

        # radial correction factor relative to peak max in first angular slice
        radcorr = radial[pkpos[0]] / radial[pkpos]

    elif method == "lsq":
        # least-squares radially scale intensity profile matching previous slice

        # initial guess fit parameters: radial correction factor, and amplitude
        fitpar = np.array([1.0, 1.0])

        # storage for the radial correction factors
        radcorr = []
        radcorr.append(1)  # first slice nothing to compare with
        previous = polarIMTrans[0]

        for ang, aslice in zip(angles[1:], polarIMTrans[1:]):
            profile = aslice

            result = leastsq(_residual, fitpar, args=(radial, profile,
                             previous))

            radcorr.append(result[0][0])  # radial scale factor direct from lsq

            previous += _residual(result[0], radial, profile, previous)
            # This "previous" slice corresponds to the previous slice intensity
            # profile that has been re-scaled. Thus, if the next slice is
            # identical, it will be assigned a scale factor of 1.0

            # use the determined radial scale factor, and amplitude parameters
            # for the next slice
            fitpar = result[0]

    else:
        raise ValueError("method variable must be one of 'argmax' or 'lsq',"
                         " not '{}'".format(method))

    return radcorr



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

import numpy as np
from numpy.testing import assert_allclose
import abel
from abel.benchmark import absolute_ratio_benchmark

DATA_DIR = os.path.join(os.path.split(__file__)[0], 'data')

def test_linbasex_shape():
    n = 21
    x = np.ones((n, n), dtype='float32')

    recon = abel.linbasex.linbasex_transform_full(x, basis_dir=None)

    assert recon[0].shape == (n+1, n+1)   # NB shape+1


def test_linbasex_forward_dribinski_image():
    """ Check hansenlaw forward/inverse transform
        using BASEX sample image, comparing speed distributions
    """

    # BASEX sample image
    IM = abel.tools.analytical.SampleImage(n=1001, name="dribinski").image

    # forward Abel transform
    fIM = abel.Transform(IM, method='hansenlaw', direction='forward')

    # inverse Abel transform
    ifIM = abel.Transform(fIM.transform, method='linbasex',
                          transform_options=dict(legendre_orders=[0, 2], 
                                            proj_angles=[0, np.pi/2],
                                            return_Beta=True))

    # speed distribution
    orig_radial, orig_speed = abel.tools.vmi.angular_integration_3D(IM)


    radial = ifIM.radial
    speed = ifIM.Beta[0]

    orig_speed /= orig_speed[1:60].max()
    speed /= speed[1:60].max()

    assert np.allclose(orig_speed[1:60], speed[1:60], rtol=0, atol=0.1)


if __name__ == "__main__":
    test_linbasex_shape()
    test_linbasex_forward_dribinski_image()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

import numpy as np
from numpy.testing import assert_allclose

import abel
from abel.basex import get_bs_cached, cache_cleanup
from abel.tools.analytical import GaussianAnalytical


DATA_DIR = os.path.join(os.path.split(__file__)[0], 'data')


def get_basis_file_name(n, sigma):
    return os.path.join(DATA_DIR,
                        'basex_basis_{}_{}.npy'.format(n, float(sigma)))


def test_basex_basis_sets_cache():
    # n = 61  (121 full width)
    # sigma = 1  (nbf = 61)
    n = 61
    sigma = 1.0
    file_name = get_basis_file_name(n, sigma)
    if os.path.exists(file_name):
        os.remove(file_name)
    # 1st call generate and save
    get_bs_cached(n, basis_dir=DATA_DIR, verbose=False)
    # 2nd call load from file
    get_bs_cached(n, basis_dir=DATA_DIR, verbose=False)
    if os.path.exists(file_name):
        os.remove(file_name)


def basex_basis_sets_resize(sigma):
    n_s = 50
    n_l = 100
    file_name_s = get_basis_file_name(n_s, sigma)
    file_name_l = get_basis_file_name(n_l, sigma)

    # (remove both basis files)
    def remove_files():
        if os.path.exists(file_name_s):
            os.remove(file_name_s)
        if os.path.exists(file_name_l):
            os.remove(file_name_l)

    # make sure that basis files do not exist and are not cached
    remove_files()
    cache_cleanup()
    # generate small basis and save
    Ai_s = get_bs_cached(n_s, sigma, correction=False,
                         basis_dir=DATA_DIR, verbose=False)
    cache_cleanup()
    # extend to large basis and save
    Ai_s_l = get_bs_cached(n_l, sigma, correction=False,
                           basis_dir=DATA_DIR, verbose=False)
    cache_cleanup()
    # delete basis files
    remove_files()
    # generate large basis and save
    Ai_l = get_bs_cached(n_l, sigma, correction=False,
                         basis_dir=DATA_DIR, verbose=False)
    cache_cleanup()
    # crop large basis to small
    Ai_l_s = get_bs_cached(n_s, sigma, correction=False,
                           basis_dir=DATA_DIR, verbose=False)
    cache_cleanup()
    # delete basis files (clean-up)
    remove_files()

    assert_allclose(Ai_s, Ai_l_s, atol=1e-15, rtol=1e-15)
    assert_allclose(Ai_l, Ai_s_l, atol=1e-15, rtol=1e-15)


def test_basex_basis_sets_resize_1():
    """Test basis resize with default sigma=1"""
    basex_basis_sets_resize(1)


def test_basex_basis_sets_resize_1_5():
    """Test basis resize with sigma=1.5
       (with n not aligned)"""
    basex_basis_sets_resize(1.5)


def test_basex_shape():
    n = 21
    x = np.ones((n, n), dtype='float32')
    Ai = abel.basex.get_bs_cached(n, basis_dir=None, verbose=False)

    recon = abel.basex.basex_core_transform(x, Ai)

    assert recon.shape == (n, n)


def test_basex_zeros():
    n = 21
    x = np.zeros((n, n), dtype='float32')
    Ai = abel.basex.get_bs_cached(n, basis_dir=None, verbose=False)

    recon = abel.basex.basex_core_transform(x, Ai)

    assert_allclose(recon, 0)


def basex_gaussian(sigma, reg, cor, tol):
    """Check a gaussian solution for BASEX"""
    n = 100
    r_max = n - 1

    ref = GaussianAnalytical(n, r_max, symmetric=False, sigma=30)
    tr = np.tile(ref.abel[None, :], (n, 1))  # make a 2D array from 1D

    correction = cor if isinstance(cor, bool) else False

    Ai = abel.basex.get_bs_cached(n, sigma=sigma, reg=reg,
                                  correction=correction,
                                  basis_dir=None, verbose=False)

    recon = abel.basex.basex_core_transform(tr, Ai)
    recon = recon[n // 2 + n % 2]

    ref = ref.func
    if cor is not True:
        # old-style intensity correction
        recon /= cor
        # skip artifact from k = 0 near r = 0
        # see https://github.com/PyAbel/PyAbel/issues/230
        cut = int(2 * sigma)
        recon = recon[cut:]
        ref = ref[cut:]

    assert_allclose(recon, ref, atol=tol)


def test_basex_gaussian():
    """Check a gaussian solution for BASEX:
       default parameters"""
    # (intensity correction using "magic number",
    #  see https://github.com/PyAbel/PyAbel/issues/230)
    basex_gaussian(sigma=1, reg=0, cor=True, tol=7e-4)


def test_basex_gaussian_uncorrected():
    """Check a gaussian solution for BASEX:
       default parameters, without correction"""
    basex_gaussian(sigma=1, reg=0, cor=1.015, tol=3e-3)


def test_basex_gaussian_sigma_3():
    """Check a gaussian solution for BASEX:
       large sigma (oscillating), without correction"""
    basex_gaussian(sigma=3, reg=0, cor=1, tol=3e-2)


def test_basex_gaussian_sigma_3_corrected():
    """Check a gaussian solution for BASEX:
       large sigma (oscillating)"""
    basex_gaussian(sigma=3, reg=0, cor=True, tol=2e-3)


def test_basex_gaussian_sigma_07_reg_10_corrected():
    """Check a gaussian solution for BASEX:
       small sigma, regularized, corrected"""
    basex_gaussian(sigma=0.7, reg=10, cor=True, tol=6e-3)


def basex_forward_gaussian(sigma, reg, atol, rtol):
    """Check a gaussian solution for BASEX"""
    n = 100
    r_max = n - 1

    ref = GaussianAnalytical(n, r_max, symmetric=False, sigma=30)
    tr = np.tile(ref.func[None, :], (n, 1))  # make a 2D array from 1D

    Ai = abel.basex.get_bs_cached(n, sigma=sigma, reg=reg,
                                  basis_dir=None, verbose=False,
                                  direction='forward')

    proj = abel.basex.basex_core_transform(tr, Ai)
    proj = proj[n // 2 + n % 2]

    ref = ref.abel

    assert_allclose(proj, ref, atol=atol, rtol=rtol)


def test_basex_forward_gaussian():
    """Check a gaussian solution for BASEX forward transform:
       default parameters"""
    basex_forward_gaussian(sigma=1, reg=0, atol=1e-3, rtol=2e-3)


def test_basex_forward_gaussian_3():
    """Check a gaussian solution for BASEX forward transform:
       large sigma"""
    basex_forward_gaussian(sigma=3, reg=0, atol=1e-3, rtol=4e-3)


def test_basex_forward_gaussian_07():
    """Check a gaussian solution for BASEX forward transform:
       small sigma, regularized"""
    basex_forward_gaussian(sigma=0.7, reg=1e-6, atol=1e-3, rtol=1e-2)


if __name__ == '__main__':
    test_basex_basis_sets_cache()
    test_basex_basis_sets_resize_1()
    test_basex_basis_sets_resize_1_5()
    test_basex_shape()
    test_basex_zeros()
    test_basex_gaussian()
    test_basex_gaussian_uncorrected()
    test_basex_gaussian_sigma_3()
    test_basex_gaussian_sigma_3_corrected()
    test_basex_gaussian_sigma_07_reg_10_corrected()
    test_basex_forward_gaussian()
    test_basex_forward_gaussian_3()
    test_basex_forward_gaussian_07()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

import numpy as np
from numpy.testing import assert_allclose, assert_array_less

from abel.rbasex import rbasex_transform, cache_cleanup
from abel.rbasex import get_bs_cached, cache_cleanup
from abel.tools.analytical import GaussianAnalytical
from abel.hansenlaw import hansenlaw_transform
from abel import Transform


DATA_DIR = os.path.join(os.path.split(__file__)[0], 'data')


def test_rbasex_shape():
    rmax = 11
    n = 2 * rmax - 1
    im = np.ones((n, n), dtype='float')

    fwd_im, fwd_distr = rbasex_transform(im, direction='forward')
    assert fwd_im.shape == (n, n)
    assert fwd_distr.r.shape == (rmax,)

    inv_im, inv_distr = rbasex_transform(im)
    assert inv_im.shape == (n, n)
    assert inv_distr.r.shape == (rmax,)


def test_rbasex_zeros():
    rmax = 11
    n = 2 * rmax - 1
    im = np.zeros((n, n), dtype='float')

    fwd_im, fwd_distr = rbasex_transform(im, direction='forward')
    assert fwd_im.shape == (n, n)
    assert fwd_distr.r.shape == (rmax,)

    inv_im, inv_distr = rbasex_transform(im)
    assert inv_im.shape == (n, n)
    assert inv_distr.r.shape == (rmax,)


def test_rbasex_gaussian():
    """Check an isotropic gaussian solution for rBasex"""
    rmax = 100
    sigma = 30
    n = 2 * rmax - 1

    ref = GaussianAnalytical(n, rmax, symmetric=True, sigma=sigma)
    # images as direct products
    src = ref.func * ref.func[:, None]
    proj = ref.abel * ref.func[:, None]  # (vertical is not Abel-transformed)

    fwd_im, fwd_distr = rbasex_transform(src, direction='forward')
    # whole image
    assert_allclose(fwd_im, proj, rtol=0.02, atol=0.001)
    # radial intensity profile (without r = 0)
    assert_allclose(fwd_distr.harmonics()[0, 1:], ref.abel[rmax:],
                    rtol=0.02, atol=5e-4)

    inv_im, inv_distr = rbasex_transform(proj)
    # whole image
    assert_allclose(inv_im, src, rtol=0.02, atol=0.02)
    # radial intensity profile (without r = 0)
    assert_allclose(inv_distr.harmonics()[0, 1:], ref.func[rmax:],
                    rtol=0.02, atol=1e-4)


def run_orders(odd=False):
    """
    Test angular orders using Gaussian peaks by comparison with hansenlaw.
    """
    maxorder = 6
    sigma = 5.0  # Gaussian sigma
    step = 6 * sigma  # distance between peak centers

    for order in range(maxorder + 1):
        rmax = int((order + 2) * step)
        if odd:
            if order == 0:
                continue  # 0th order cannot be odd
            height = 2 * rmax + 1
        else:  # even only
            if order % 2:
                continue  # skip odd
            height = rmax + 1
        # coordinates (Q0 or right half):
        x = np.arange(float(rmax + 1))
        y = rmax - np.arange(float(height))[:, None]
        # radius
        r = np.sqrt(x**2 + y**2)
        # cos, sin
        r[rmax, 0] = np.inf
        c = y / r
        s = x / r
        r[rmax, 0] = 0

        # Gaussian peak with one cossin angular term
        def peak(i):
            m = i  # cos power
            k = (order - m) & ~1  # sin power (round down to even)
            return c ** m * s ** k * \
                   np.exp(-(r - (i + 1) * step) ** 2 / (2 * sigma**2))

        # create source distribution
        src = peak(0)
        for i in range(1, order + 1):
            if not odd and i % 2:
                continue  # skip odd
            src += peak(i)

        # reference forward transform
        abel = hansenlaw_transform(src, direction='forward', hold_order=1)

        param = ', order = {}, odd = {}, '.format(order, odd)

        # test forward transform
        for mode in ['clean', 'cached']:
            if mode == 'clean':
                cache_cleanup()
            proj, _ = rbasex_transform(src, origin=(rmax, 0),
                                       order=order, odd=odd,
                                       direction='forward', out='fold')
            assert_allclose(proj, abel, rtol=0.003, atol=0.4,
                            err_msg='-> forward' + param + mode)

        # test inverse transforms
        for reg in [None, ('L2', 1), ('diff', 1), ('SVD', 1 / rmax)]:
            for mode in ['clean', 'cached']:
                if mode == 'clean':
                    cache_cleanup()
                recon, _ = rbasex_transform(abel, origin=(rmax, 0),
                                            order=order, odd=odd,
                                            reg=reg, out='fold')
                recon[rmax-2:rmax+3, :2] = 0  # exclude pixels near center
                assert_allclose(recon, src, atol=0.03,
                                err_msg='-> reg = ' + str(reg) + param + mode)


def test_rbasex_orders():
    run_orders()


def test_rbasex_orders_odd():
    run_orders(odd=True)


def test_rbasex_pos():
    """
    Test positive regularization as in run_orders().
    """
    sigma = 5.0  # Gaussian sigma
    step = 6 * sigma  # distance between peak centers

    for order in [0, 1, 2, 4, 6]:
        rmax = int((order + 2) * step)
        if order == 1:  # odd
            rmax += int(step)  # 3 peaks instead of 2
            height = 2 * rmax + 1
        else:  # even only
            height = rmax + 1
        # coordinates (Q0 or right half):
        x = np.arange(float(rmax + 1))
        y = rmax - np.arange(float(height))[:, None]
        # radius
        r = np.sqrt(x**2 + y**2)
        # cos, sin
        r[rmax, 0] = np.inf
        c = y / r
        s = x / r
        r[rmax, 0] = 0

        # Gaussian peak with one cossin angular term
        def peak(i, isotropic=False):
            if isotropic:
                return np.exp(-(r - (i + 1) * step) ** 2 / (2 * sigma**2))
            m = i  # cos power
            k = (order - m) & ~1  # sin power (round down to even)
            return c ** m * s ** k * \
                   np.exp(-(r - (i + 1) * step) ** 2 / (2 * sigma**2))

        # create source distribution
        src = peak(0)
        if order == 1:  # special case
            src += (1 + c) * peak(1, True)  # 1 + cos >= 0
            src += (1 - c) * peak(2, True)  # 1 - cos >= 0
        else:  # other even orders
            for i in range(2, order + 1, 2):
                src += peak(i)
        # array for nonnegativity test (with some tolerance)
        zero = np.full_like(src, -1e-15)

        # reference forward transform
        abel = hansenlaw_transform(src, direction='forward', hold_order=1)
        # with some noise
        abel += 0.05 * np.random.RandomState(0).rand(*abel.shape)

        param = '-> order = {}, '.format(order)

        # test inverse transform
        for mode in ['clean', 'cached']:
            if mode == 'clean':
                cache_cleanup()
            recon, _ = rbasex_transform(abel, origin=(rmax, 0),
                                        order=order,
                                        reg='pos', out='fold')
            recon[rmax-3:rmax+4, :2] = 0  # exclude pixels near center
            assert_allclose(recon, src, atol=0.05,
                            err_msg=param + mode)
            # check nonnegativity
            assert_array_less(zero, recon)


def run_out(odd=False):
    """
    Test some output shapes.
    """
    sigma = 5.0  # Gaussian sigma
    step = int(6 * sigma)  # distance between peak centers

    rmax = (2 + 2) * step
    size = 2 * rmax + 1
    # coordinates (full image):
    x = np.arange(float(size)) - rmax
    y = rmax - np.arange(float(size))[:, None]
    # radius
    r = np.sqrt(x**2 + y**2)
    # cos
    r[rmax, rmax] = np.inf
    c = y / r
    r[rmax, rmax] = 0

    # Gaussian peak with one cos^n angular term
    def peak(i):
        n = i if odd else 2 * i
        return c ** n * \
               np.exp(-(r - (i + 1) * step) ** 2 / (2 * sigma**2))

    # create source distribution
    src = peak(0) + peak(1)

    # reference forward transform
    abel = Transform(src, direction='forward', method='hansenlaw',
                     transform_options={'hold_order': 1}).transform

    param = '-> odd = {}, '.format(odd)

    # Test forward transform:

    # rmax < MIN => larger 'same'
    proj, _ = rbasex_transform(src, rmax=rmax - step, odd=odd,
                               direction='forward', out='same')
    assert_allclose(proj, abel, rtol=0.005, atol=0.2,
                    err_msg=param + 'rmax < MIN, out = same')

    # cropped, rmax = HOR < VER -> same
    crop = (slice(3 * step, -step // 2),
            slice(3 * step, -step))
    proj, _ = rbasex_transform(src[crop], origin=(step, step),
                               odd=odd, direction='forward', out='same')
    assert_allclose(proj, abel[crop], rtol=0.005, atol=0.2,
                    err_msg=param + 'rmax = HOR < VER, out = same')

    # cropped, rmax = VER < HOR -> same
    crop = (slice(step, -3 * step),
            slice(3 * step, -step // 2))
    proj, _ = rbasex_transform(src[crop], origin=(3 * step, step),
                               odd=odd, direction='forward', out='same')
    assert_allclose(proj, abel[crop], rtol=0.0003, atol=0.3,
                    err_msg=param + 'rmax = VER < HOR, out = same')

    # cropped, rmax = VER > HOR -> same
    crop = (slice(3 * step, -step // 2),
            slice(3 * step, -step))
    proj, _ = rbasex_transform(src[crop], origin=(step, step), rmax='MAX',
                               odd=odd, direction='forward', out='same')
    assert_allclose(proj, abel[crop], rtol=0.005, atol=0.2,
                    err_msg=param + 'rmax = VER > HOR, out = same')

    # cropped, rmax = HOR > VER -> same
    crop = (slice(step, -3 * step),
            slice(3 * step, -step // 2))
    proj, _ = rbasex_transform(src[crop], origin=(3 * step, step), rmax='MAX',
                               odd=odd, direction='forward', out='same')
    assert_allclose(proj, abel[crop], rtol=0.0003, atol=0.3,
                    err_msg=param + 'rmax = HOR > VER, out = same')

    # cropped, rmax = rmax -> full
    crop = (slice(3 * step, -step),
            slice(3 * step, -step))
    # (in multiples of step: VER^2 + HOR^2 = 2 * 3^2 > rmax^2 = 4^2)
    proj, _ = rbasex_transform(src[crop], origin=(step, step), rmax=rmax,
                               odd=odd, direction='forward', out='full')
    assert_allclose(proj, abel, rtol=0.002, atol=0.3,
                    err_msg=param + 'rmax = rmax, out = full')


def test_rbasex_out():
    run_out()


def test_rbasex_out_odd():
    run_out(odd=True)


def get_basis_file_name(rmax, order, odd, inv):
    return os.path.join(DATA_DIR,
                        'rbasex_basis_{}_{}{}{}.npy'.format(rmax, order,
                                                            'o' if odd else '',
                                                            'i' if inv else ''))


def test_rbasex_bs_cache():
    rmax, order, odd, inv = 50, 2, False, True
    file_name = get_basis_file_name(rmax, order, odd, inv)
    if os.path.exists(file_name):
        os.remove(file_name)
    # 1st call generate and save
    bs1 = get_bs_cached(rmax, order, odd, basis_dir=DATA_DIR, verbose=True)
    assert os.path.exists(file_name), 'Basis set was not saved!'
    # 2nd call load from file
    bs2 = get_bs_cached(rmax, order, odd, basis_dir=DATA_DIR, verbose=True)
    assert_allclose(bs1, bs2, err_msg='Loaded basis set differs from saved!')
    if os.path.exists(file_name):
        os.remove(file_name)


def rbasex_bs_resize(old, new, new_saved=False):
    """
    Compare basis set loaded and modified to cleanly computed.
    old, new: basis parameters (rmax, order, odd, inv)
    new_saved: check that new basis is (True) or is not (False) saved
    """
    file_name_old = get_basis_file_name(*old)
    file_name_new = get_basis_file_name(*new)

    # (remove both basis files)
    def remove_files():
        if os.path.exists(file_name_old):
            os.remove(file_name_old)
        if os.path.exists(file_name_new):
            os.remove(file_name_new)

    # (convert basis-set parameters to get_bs_cached() arguments)
    def arg(prm):
        return prm[:-1] + ('inverse' if prm[3] else 'forward',)

    # make sure that basis files do not exist and are not cached
    remove_files()
    cache_cleanup()
    # generate "old" basis and save
    get_bs_cached(*arg(old), basis_dir=DATA_DIR, verbose=False)
    cache_cleanup()
    # generate "new" basis without saving
    bs_new = get_bs_cached(*arg(new), basis_dir=None, verbose=False)
    cache_cleanup()

    # load "new" basis from "old" file, cropped
    bs = get_bs_cached(*arg(new), basis_dir=DATA_DIR, verbose=True)
    # check that "new" file is (not) be saved
    if new_saved:
        assert os.path.exists(file_name_new), "New basis set was not saved!"
    else:
        assert not os.path.exists(file_name_new), "New basis set was saved!"
    # clean-up all caches
    cache_cleanup()
    remove_files()

    # compare cropped to clean
    assert_allclose(bs, bs_new)


def test_rbasex_bs_crop_rmax():
    rbasex_bs_resize((60, 2, False, True),
                     (50, 2, False, True))


def test_rbasex_bs_crop_order():
    rbasex_bs_resize((50, 2, False, True),
                     (50, 0, False, True))


def test_rbasex_bs_crop_odd():
    rbasex_bs_resize((50, 2, True, True),
                     (50, 2, False, True))


def test_rbasex_bs_crop_order_odd():
    rbasex_bs_resize((50, 3, True, True),
                     (50, 2, False, True))


def test_rbasex_bs_crop_inv():
    rbasex_bs_resize((50, 2, False, True),
                     (50, 2, False, False))


def test_rbasex_bs_add_inv():
    rbasex_bs_resize((50, 2, False, False),
                     (50, 2, False, True),
                     new_saved=True)


if __name__ == '__main__':
    test_rbasex_shape()
    test_rbasex_zeros()
    test_rbasex_gaussian()
    test_rbasex_orders()
    test_rbasex_orders_odd()
    test_rbasex_pos()
    test_rbasex_out()
    test_rbasex_out_odd()
    test_rbasex_bs_cache()
    test_rbasex_bs_crop_rmax()
    test_rbasex_bs_crop_order()
    test_rbasex_bs_crop_odd()
    test_rbasex_bs_crop_order_odd()
    test_rbasex_bs_crop_inv()
    test_rbasex_bs_add_inv()



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np
from numpy.testing import assert_allclose
from scipy.special import hermite
from math import factorial

import abel
from abel.tools.polynomial import Polynomial, PiecewisePolynomial


def test_polynomial_shape():
    """
    Testing that func and abel have the same shape as r.
    """
    n = 20
    r = np.linspace(0, n/2, n)
    P = Polynomial(r, 0, n, [1])

    assert P.func.shape == r.shape
    assert P.abel.shape == r.shape


def test_polynomial_zeros():
    """
    Testing that zero coefficients produce zero func and abel.
    """
    n = 20
    r = np.linspace(0, n/2, n)
    P = Polynomial(r, 0, n, [0, 0])

    assert_allclose(P.func, 0)
    assert_allclose(P.abel, 0)


def test_polynomial_step():
    """
    Testing step function (and multiplication).
    """
    n = 50
    r = np.linspace(0, n/2, n)
    r_min = 10
    r_max = 20
    A = 2

    P = A * Polynomial(r, r_min, r_max, [1/A])

    func = np.zeros_like(r)
    func[(r >= r_min) * (r <= r_max)] = 1

    def sqrt0(x): return np.sqrt(x, np.zeros_like(x), where=x > 0)
    abel = 2 * (sqrt0(r_max**2 - r**2) - sqrt0(r_min**2 - r**2))

    assert_allclose(P.func, func)
    assert_allclose(P.abel, abel)


def test_polynomial_gaussian():
    """
    Testing shifted and scaled Taylor expansion of a Gaussian,
    in reduced coordinates.
    """
    n = 100
    r = np.linspace(0, n/2, n)
    sigma = 15

    # coefficients of Taylor series around r = 1
    c = []
    for k in range(50):
        c.append(np.exp(-1.0) * hermite(k)(-1.0) / factorial(k))

    P = Polynomial(r, 0, r[-1] + 2, c, sigma, sigma, reduced=True)
    # scaling and shifting by sigma should produce exp(-(r/σ)^2),
    # its Abel transform is √π σ exp(-(r/σ)^2)

    func = np.exp(-(r / sigma)**2)
    abel = np.sqrt(np.pi) * sigma * func

    assert_allclose(P.func, func, atol=1.0e-7)
    assert_allclose(P.abel, abel, atol=1.0e-4)


def test_polynomial_smoothstep():
    """
    Testing cubic smoothstep function,
    using ``direct`` inverse transform.
    """
    n = 100
    r = np.arange(float(n))
    r_min = 20
    r_max = 80
    w = 10
    h = 3

    c = [1/2, 3/4, 0, -1/4]
    P = PiecewisePolynomial(r, [(r_min - w, r_min + w, c, r_min, w),
                                (r_min + w, r_max - w, [1]),
                                (r_max - w, r_max + w, c, r_max, -w)])
    P *= h

    recon = abel.direct.direct_transform(P.abel, backend='python')

    assert_allclose(P.func, recon, atol=2.0e-2)


if __name__ == '__main__':
    test_polynomial_shape()
    test_polynomial_zeros()
    test_polynomial_step()
    test_polynomial_gaussian()
    test_polynomial_smoothstep()



# -*- coding: utf-8 -*-
from __future__ import division

import numpy as np
from scipy.linalg import pascal, toeplitz

__doc__ = """
See :ref:`Polynomials` for details and examples.

.. toctree::
    :hidden:

    tools/polynomial
"""


class Polynomial(object):
    r"""
    Polynomial function and its Abel transform.

    Supports multiplication and division by numbers.

    Parameters
    ----------
    r : numpy array
        *r* values at which the function is generated
        (and *x* values for its Abel transform);
        must be non-negative and in ascending order
    r_min, r_max : float
        *r* domain:
        the function is defined as the polynomial on [**r_min**, **r_max**]
        and zero outside it;
        0 ≤ **r_min** < **r_max** ≲ **max r**
        (**r_max** might exceed maximal **r**, but usually by < 1 pixel;
        negative **r_min** or **r_max** are allowed for convenience but are
        interpreted as 0)
    c: numpy array
        polynomial coefficients in order of increasing degree:
        [c₀, c₁, c₂] means c₀ + c₁ *r* + c₂ *r*\ ²
    r_0 : float, optional
        origin shift: the polynomial is defined as
        c₀ + c₁ (*r* − **r_0**) + c₂ (*r* − **r_0**)² + ...
    s : float, optional
        *r* stretching factor (around **r_0**): the polynomial is defined as
        c₀ + c₁ (*r*/**s**) + c₂ (*r*/**s**)² + ...
    reduced : boolean, optional
        internally rescale the *r* range to [0, 1];
        useful to avoid floating-point overflows for high degrees
        at large r (and might improve numeric accuracy)
    """
    def __init__(self, r, r_min, r_max, c, r_0=0.0, s=1.0, reduced=False):
        n = r.shape[0]

        # trim negative r limits
        if r_max <= 0:
            # both func and abel must be zero everywhere
            self.func = np.zeros(n)
            self.abel = np.zeros(n)
            return
        if r_min < 0:
            r_min = 0

        # remove zero high-order terms
        c = np.array(np.trim_zeros(c, 'b'), float)
        # if all coefficients are zero
        if len(c) == 0:
            # then both func and abel are also zero everywhere
            self.func = np.zeros(n)
            self.abel = np.zeros(n)
            return
        # polynomial degree
        K = len(c) - 1

        if reduced:
            # rescale r to [0, 1] (to avoid FP overflow)
            r = r / r_max
            r_0 /= r_max
            s /= r_max
            abel_scale = r_max
            r_min /= r_max
            r_max = 1.0

        if s != 1.0:
            # apply stretch
            S = np.cumprod([1.0] + [1.0 / s] * K)  # powers of 1/s
            c *= S
        if r_0 != 0.0:
            # apply shift
            P = pascal(1 + K, 'upper', False)  # binomial coefficients
            rk = np.cumprod([1.0] + [-float(r_0)] * K)  # powers of -r_0
            T = toeplitz([1.0] + [0.0] * K, rk)  # upper-diag. (-r_0)^{l - k}
            c = (P * T).dot(c)

        # whether even and odd powers are present
        even = np.any(c[::2])
        odd = np.any(c[1::2])

        # index limits
        i_min = np.searchsorted(r, r_min)
        i_max = np.searchsorted(r, r_max)

        # Calculate all necessary variables within [0, r_max]

        # x, x^2
        x = r[:i_max]
        x2 = x * x

        # integration limits y = sqrt(r^2 - x^2) or 0
        def sqrt0(x): return np.sqrt(x, np.zeros_like(x), where=x > 0)
        y_up = sqrt0(r_max * r_max - x2)
        y_lo = sqrt0(r_min * r_min - x2)

        # y r^k |_lo^up
        # (actually only even are neded for "even", and only odd for "odd")
        Dyr = np.outer(np.cumprod([1.0] + [r_max] * K), y_up) - \
              np.outer(np.cumprod([1.0] + [r_min] * K), y_lo)

        # ln(r + y) |_lo^up, only for odd k
        if odd:
            # ln x for x > 0, otherwise 0
            def ln0(x): return np.log(x, np.zeros_like(x), where=x > 0)
            Dlnry = ln0(r_max + y_up) - \
                    ln0(np.maximum(r_min, x) + y_lo)

        # One-sided Abel integral \int_lo^up r^k dy.
        def a(k):
            odd_k = k % 2
            # max. x power
            K = k - odd_k  # (k - 1 for odd k)
            # generate coefficients for all x^m r^{k - m} terms
            # (only even indices are actually used;
            #  for odd k, C[K] is also used for x^{k+1} ln(r + y))
            C = [0] * (K + 1)
            C[0] = 1 / (k + 1)
            for m in range(k, 1, -2):
                C[k - m + 2] = C[k - m] * m / (m - 1)
            # sum all terms using Horner's method in x
            a = C[K] * Dyr[k - K]
            if odd_k:
                a += C[K] * x2 * Dlnry
            for m in range(K - 2, -1, -2):
                a = a * x2 + C[m] * Dyr[k - m]
            return a

        # Generate the polynomial function
        func = np.zeros(n)
        span = slice(i_min, i_max)
        # (using Horner's method)
        func[span] = c[K]
        for k in range(K - 1, -1, -1):
            func[span] = func[span] * x[span] + c[k]
        self.func = func

        # Generate its Abel transform
        abel = np.zeros(n)
        span = slice(0, i_max)
        if reduced:
            c *= abel_scale
        for k in range(K + 1):
            if c[k]:
                abel[span] += c[k] * 2 * a(k)
        self.abel = abel

    def __imul__(self, m):
        """
        In-place multiplication: Polynomial *= num.
        """
        self.func *= m
        self.abel *= m
        return self

    def __mul__(self, m):
        """
        Multiplication: Polynomial * num.
        """
        res = self.__new__(type(self))  # create empty object (same type)
        res.func = self.func * m
        res.abel = self.abel * m
        return res

    __rmul__ = __mul__
    __rmul__.__doc__ = \
        """
        Multiplication: num * Polynomial.
        """

    def __itruediv__(self, m):
        """
        In-place division: Polynomial /= num.
        """
        return self.__imul__(1 / m)

    def __truediv__(self, m):
        """
        Division: Polynomial / num.
        """
        return self.__mul__(1 / m)

    # (Addition and subtraction are not implemented because they are
    #  meaningful only for polynomials generated over the same r array.
    #  Use PiecewisePolynomial for sums of polynomials.)


class PiecewisePolynomial(Polynomial):
    r"""
    Piecewise polynomial function (sum of ``Polynomial``\ s)
    and its Abel transform.

    Supports multiplication and division by numbers.

    Parameters
    ----------
    r : numpy array
        *r* values at which the function is generated
        (and *x* values for its Abel transform)
    ranges : iterable of unpackable
        (list of tuples of) polynomial parameters for each piece::

           [(r_min_1st, r_max_1st, c_1st),
            (r_min_2nd, r_max_2nd, c_2nd),
            ...
            (r_min_nth, r_max_nth, c_nth)]

        according to ``Polynomial`` conventions.
        All ranges are independent (may overlap and have gaps, may define
        polynomials of any degrees) and may include optional ``Polynomial``
        parameters
    """
    def __init__(self, r, ranges):
        self.p = []
        for rng in ranges:
            self.p.append(Polynomial(r, *rng))

        func = self.p[0].func
        for p in self.p[1:]:
            func += p.func
        self.func = func

        abel = self.p[0].abel
        for p in self.p[1:]:
            abel += p.abel
        self.abel = abel

    # (Multiplication and division methods are inherited from Polynomial.)
